# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging
import os
import pathlib
import shutil
import tempfile
import zipfile

import geojson
import requests
from lxml import html
from shapely import geometry

from tests import EODagTestCase
from tests.context import (
    DEFAULT_SHAPELY_GEOMETRY,
    DEFAULT_STREAM_REQUESTS_TIMEOUT,
    NOT_AVAILABLE,
    USER_AGENT,
    DatasetDriver,
    Download,
    EOProduct,
    HTTPDownload,
    ProgressCallback,
    config,
)
from tests.utils import mock


class TestEOProduct(EODagTestCase):
    NOT_ASSOCIATED_PRODUCT_TYPE = "EODAG_DOES_NOT_SUPPORT_THIS_PRODUCT_TYPE"

    def setUp(self):
        super(TestEOProduct, self).setUp()
        self.output_dir = tempfile.mkdtemp()

    def tearDown(self):
        super(TestEOProduct, self).tearDown()
        if os.path.isdir(self.output_dir):
            shutil.rmtree(self.output_dir)

    def get_mock_downloader(self):
        """Returns a mock downloader with a default configuration."""
        mock_downloader = mock.MagicMock(
            spec_set=Download(provider=self.provider, config=None)
        )
        mock_downloader.config = config.PluginConfig.from_mapping(
            {"type": "Foo", "output_dir": tempfile.gettempdir()}
        )
        return mock_downloader

    def test_eoproduct_search_intersection_geom(self):
        """EOProduct search_intersection attr must be it's geom when no bbox_or_intersect param given"""
        product = self._dummy_product()
        self.assertEqual(product.geometry, product.search_intersection)

    def test_eoproduct_default_geom(self):
        """EOProduct needs a geometry or can use confired defaultGeometry by default"""

        product_no_default_geom = self._dummy_product(
            properties={"geometry": NOT_AVAILABLE}
        )
        self.assertEqual(product_no_default_geom.geometry, DEFAULT_SHAPELY_GEOMETRY)

        product_default_geom = self._dummy_product(
            properties={"geometry": NOT_AVAILABLE, "defaultGeometry": (0, 0, 1, 1)}
        )
        self.assertEqual(product_default_geom.geometry.bounds, (0.0, 0.0, 1.0, 1.0))

    def test_eoproduct_search_intersection_none(self):
        """EOProduct search_intersection attr must be None if shapely.errors.GEOSException when intersecting"""
        # Invalid geometry
        self.eoproduct_props["geometry"] = {
            "type": "Polygon",
            "coordinates": [
                [
                    [10.469970703124998, 3.9957805129630373],
                    [12.227783203124998, 4.740675384778385],
                    [12.095947265625, 4.061535597066097],
                    [10.491943359375, 4.412136788910175],
                    [10.469970703124998, 3.9957805129630373],
                ]
            ],
        }
        product = self._dummy_product(
            geometry=geometry.Polygon(
                (
                    (10.469970703124998, 3.9957805129630373),
                    (10.469970703124998, 4.740675384778385),
                    (12.227783203124998, 4.740675384778385),
                    (12.227783203124998, 3.9957805129630373),
                )
            ),
        )
        self.assertIsNone(product.search_intersection)

    def test_eoproduct_default_driver_unsupported_product_type(self):
        """EOProduct driver attr must be set even if its product type is not supported"""
        product = self._dummy_product(productType=self.NOT_ASSOCIATED_PRODUCT_TYPE)
        self.assertIsInstance(product.driver, DatasetDriver)

    def test_eoproduct_geointerface(self):
        """EOProduct must provide a geo-interface with a set of specific properties"""
        product = self._dummy_product()
        geo_interface = geojson.loads(geojson.dumps(product))
        self.assertEqual(geo_interface["type"], "Feature")
        self.assertEqual(
            geo_interface["geometry"],
            self._tuples_to_lists(geometry.mapping(self.geometry)),
        )
        properties = geo_interface["properties"]
        self.assertEqual(properties["eodag_provider"], self.provider)
        self.assertEqual(
            properties["eodag_search_intersection"],
            self._tuples_to_lists(geometry.mapping(product.search_intersection)),
        )
        self.assertEqual(properties["eodag_product_type"], self.product_type)

    def test_eoproduct_from_geointerface(self):
        """EOProduct must be build-able from its geo-interface"""
        product = self._dummy_product()
        same_product = EOProduct.from_geojson(geojson.loads(geojson.dumps(product)))
        self.assertSequenceEqual(
            [
                product.provider,
                product.location,
                product.properties["title"],
                product.properties["instrument"],
                self._tuples_to_lists(geometry.mapping(product.geometry)),
                self._tuples_to_lists(geometry.mapping(product.search_intersection)),
                product.product_type,
                product.properties["productType"],
                product.properties["platformSerialIdentifier"],
            ],
            [
                same_product.provider,
                same_product.location,
                same_product.properties["title"],
                same_product.properties["instrument"],
                self._tuples_to_lists(geometry.mapping(same_product.geometry)),
                self._tuples_to_lists(
                    geometry.mapping(same_product.search_intersection)
                ),
                same_product.product_type,
                same_product.properties["productType"],
                same_product.properties["platformSerialIdentifier"],
            ],
        )

    def test_eoproduct_get_quicklook_no_quicklook_url(self):
        """EOProduct.get_quicklook must return an empty string if no quicklook property"""  # noqa
        product = self._dummy_product()
        product.properties["quicklook"] = None

        quicklook_file_path = product.get_quicklook()
        self.assertEqual(quicklook_file_path, "")

    def test_eoproduct_get_quicklook_http_error(self):
        """EOProduct.get_quicklook must return an empty string if there was an error during retrieval"""  # noqa
        product = self._dummy_product()
        product.properties["quicklook"] = "https://fake.url.to/quicklook"

        self.requests_http_get.return_value.__enter__.return_value.raise_for_status.side_effect = (  # noqa
            requests.HTTPError
        )
        product.register_downloader(self.get_mock_downloader(), None)

        quicklook_file_path = product.get_quicklook()
        self.assertEqual(self.requests_http_get.call_count, 2)
        self.requests_http_get.assert_called_with(
            "https://fake.url.to/quicklook",
            stream=True,
            auth=None,
            headers=USER_AGENT,
            timeout=DEFAULT_STREAM_REQUESTS_TIMEOUT,
            verify=True,
        )
        self.assertEqual(quicklook_file_path, "")

    def test_eoproduct_get_quicklook_ok_without_auth(self):
        """EOProduct.get_quicklook must retrieve the quicklook without authentication."""  # noqa
        product = self._dummy_product()
        product.properties["quicklook"] = "https://fake.url.to/quicklook"

        self.requests_http_get.return_value.__enter__.return_value.raise_for_status.side_effect = (  # noqa
            requests.HTTPError,
            None,
        )
        product.register_downloader(self.get_mock_downloader(), None)

        quicklook_file_path = product.get_quicklook()
        self.assertEqual(self.requests_http_get.call_count, 2)
        self.requests_http_get.assert_called_with(
            "https://fake.url.to/quicklook",
            stream=True,
            auth=None,
            headers=USER_AGENT,
            timeout=DEFAULT_STREAM_REQUESTS_TIMEOUT,
            verify=True,
        )
        os.remove(quicklook_file_path)

    def test_eoproduct_get_quicklook_ok(self):
        """EOProduct.get_quicklook must return the path to the successfully downloaded quicklook"""  # noqa
        product = self._dummy_product()
        product.properties["quicklook"] = "https://fake.url.to/quicklook"

        self.requests_http_get.return_value = self._quicklook_response()
        product.register_downloader(self.get_mock_downloader(), None)

        quicklook_file_path = product.get_quicklook()
        self.requests_http_get.assert_called_once_with(
            "https://fake.url.to/quicklook",
            stream=True,
            auth=None,
            headers=USER_AGENT,
            timeout=DEFAULT_STREAM_REQUESTS_TIMEOUT,
            verify=True,
        )
        self.assertEqual(
            os.path.basename(quicklook_file_path), product.properties["id"]
        )
        self.assertEqual(
            os.path.dirname(quicklook_file_path),
            os.path.join(tempfile.gettempdir(), "quicklooks"),
        )
        os.remove(quicklook_file_path)

        # Test the same thing as above but with an explicit name given to the downloaded File
        quicklook_file_path = product.get_quicklook(filename="the_quicklook.png")
        self.requests_http_get.assert_called_with(
            "https://fake.url.to/quicklook",
            stream=True,
            auth=None,
            headers=USER_AGENT,
            timeout=DEFAULT_STREAM_REQUESTS_TIMEOUT,
            verify=True,
        )
        self.assertEqual(self.requests_http_get.call_count, 2)
        self.assertEqual(os.path.basename(quicklook_file_path), "the_quicklook.png")
        self.assertEqual(
            os.path.dirname(quicklook_file_path),
            os.path.join(tempfile.gettempdir(), "quicklooks"),
        )
        os.remove(quicklook_file_path)

        # Overall teardown
        os.rmdir(os.path.dirname(quicklook_file_path))

    def test_eoproduct_get_quicklook_ok_existing(self):
        """EOProduct.get_quicklook must return the path to an already downloaded quicklook"""  # noqa
        quicklook_dir = os.path.join(tempfile.gettempdir(), "quicklooks")
        quicklook_basename = "the_quicklook.png"
        existing_quicklook_file_path = os.path.join(quicklook_dir, quicklook_basename)
        if not os.path.exists(quicklook_dir):
            os.mkdir(quicklook_dir)
        with open(existing_quicklook_file_path, "wb") as fh:
            fh.write(b"content")
        product = self._dummy_product()
        product.properties["quicklook"] = "https://fake.url.to/quicklook"
        product.register_downloader(self.get_mock_downloader(), None)

        quicklook_file_path = product.get_quicklook(filename=quicklook_basename)
        self.assertEqual(self.requests_http_get.call_count, 0)
        self.assertEqual(quicklook_file_path, existing_quicklook_file_path)
        os.remove(existing_quicklook_file_path)
        os.rmdir(quicklook_dir)

    @staticmethod
    def _quicklook_response():
        class Response(object):
            """Emulation of a response to requests.get method for a quicklook"""

            def __init__(response):
                response.headers = {"content-length": 2**5}

            def __enter__(response):
                return response

            def __exit__(response, *args):
                pass

            @staticmethod
            def iter_content(**kwargs):
                with io.BytesIO(b"a" * 2**5) as fh:
                    while True:
                        chunk = fh.read(kwargs["chunk_size"])
                        if not chunk:
                            break
                        yield chunk

            def raise_for_status(response):
                pass

        return Response()

    def test_eoproduct_download_http_default(self):
        """eoproduct.download must save the product at output_dir and create a .downloaded dir"""  # noqa
        # Setup
        product = self._dummy_downloadable_product()
        with self.assertLogs(level="INFO") as cm:
            # Download
            product_dir_path = product.download()
            self.addCleanup(self._clean_product, product_dir_path)
            self.assertIn("Download url: %s" % product.remote_location, str(cm.output))
            self.assertIn(
                "Remote location of the product is still available", str(cm.output)
            )

        # Check that the mocked request was properly called.
        self.requests_request.assert_called_once()
        download_records_dir = pathlib.Path(product_dir_path).parent / ".downloaded"
        # A .downloaded folder should be created, including a text file that
        # lists the downloaded product by their url
        self.assertTrue(download_records_dir.is_dir())
        files_in_records_dir = list(download_records_dir.iterdir())
        self.assertEqual(len(files_in_records_dir), 1)
        records_file = files_in_records_dir[0]
        actual_download_url = records_file.read_text()
        self.assertEqual(actual_download_url, self.download_url)
        # Since extraction is True by default, check that the returned path is the
        # product's directory.
        self.assertTrue(os.path.isdir(product_dir_path))
        # Check that the ZIP file is still there
        product_dir_path = pathlib.Path(product_dir_path)
        product_zip = product_dir_path.parent / (product_dir_path.name + ".zip")
        self.assertTrue(zipfile.is_zipfile(product_zip))
        # check that product is not downloaded again
        with self.assertLogs(level="INFO") as cm:
            product.download()
            self.assertIn("Product already present on this platform", str(cm.output))
        # check that product is not downloaded again even if location has not been updated
        product.location = product.remote_location
        with self.assertLogs(level="INFO") as cm:
            product.download()
            self.assertIn("Product already downloaded", str(cm.output))
            self.assertIn(
                "Extraction cancelled, destination directory already exists",
                str(cm.output),
            )

    def test_eoproduct_download_http_delete_archive(self):
        """eoproduct.download must delete the downloaded archive"""  # noqa
        # Setup
        product = self._dummy_downloadable_product(delete_archive=True)
        try:
            # Download
            product_dir_path = product.download()
            # Check that the mocked request was properly called.
            self.requests_request.assert_called_once()
            # Check that the product's directory exists.
            self.assertTrue(os.path.isdir(product_dir_path))
            # Check that the ZIP file was deleted there
            _product_dir_path = pathlib.Path(product_dir_path)
            product_zip = _product_dir_path.parent / (_product_dir_path.name + ".zip")
            self.assertFalse(os.path.exists(product_zip))
            # check that product is not downloaded again even if location has not been updated
            product.location = product.remote_location
            with self.assertLogs(level="INFO") as cm:
                product.download()
                self.assertIn("Product already downloaded", str(cm.output))
                self.assertIn(
                    "Extraction cancelled, destination directory already exists",
                    str(cm.output),
                )
        finally:
            # Teardown
            self._clean_product(product_dir_path)

    def test_eoproduct_download_http_extract(self):
        """eoproduct.download over must be able to extract a product"""
        # Setup
        product = self._dummy_downloadable_product(extract=True)
        product_dir_path = product.download()
        self.addCleanup(self._clean_product, product_dir_path)
        product_dir_path = pathlib.Path(product_dir_path)
        # The returned path must be a directory.
        self.assertTrue(product_dir_path.is_dir())
        # Check that the extracted dir has at least one file, there are more
        # but that should be enough.
        self.assertGreaterEqual(len(list(product_dir_path.glob("**/*"))), 1)
        # The zip file should be around
        product_zip_file = product_dir_path.with_suffix(".zip")
        self.assertTrue(product_zip_file.is_file)

    # TODO: add a test on tarfiles extraction

    def test_eoproduct_download_http_dynamic_options(self):
        """eoproduct.download must accept the download options to be set automatically"""
        # Setup
        product = self._dummy_product()
        self._set_download_simulation()
        dl_config = config.PluginConfig.from_mapping(
            {
                "type": "HTTPDownload",
                "base_uri": "fake_base_uri",
                "output_dir": "will_be_overriden",
            }
        )
        downloader = HTTPDownload(provider=self.provider, config=dl_config)
        product.register_downloader(downloader, None)

        output_dir_name = "_testeodag"
        output_dir = pathlib.Path(tempfile.gettempdir()) / output_dir_name
        try:
            if output_dir.is_dir():
                shutil.rmtree(output_dir)
            output_dir.mkdir()

            # Download
            product_dir_path = product.download(
                output_dir=str(output_dir),
                extract=True,
                dl_url_params={"fakeparam": "dummy"},
            )
            # Check that dl_url_params are properly passed to the GET request
            self.requests_request.assert_called_once()
            # Check that "output_dir" is respected.
            product_dir_path = pathlib.Path(product_dir_path)
            self.assertEqual(product_dir_path.parent.name, output_dir_name)
            # We've asked to extract the product so there should be a directory.
            self.assertTrue(product_dir_path.is_dir())
            # Check that the extracted dir has at least one file, there are more
            # but that should be enough.
            self.assertGreaterEqual(len(list(product_dir_path.glob("**/*"))), 1)
            # The downloaded zip file is still around
            product_zip_file = product_dir_path.with_suffix(".zip")
            self.assertTrue(product_zip_file.is_file)
        finally:
            # Teardown (all the created files are within output_dir)
            shutil.rmtree(output_dir)

    def test_eoproduct_download_progress_bar(self):
        """eoproduct.download must show a progress bar"""
        product = self._dummy_downloadable_product()
        product.properties["id"] = 12345
        progress_callback = ProgressCallback()

        # progress bar did not start
        self.assertEqual(progress_callback.n, 0)

        # extract=true would replace bar desc with extraction status
        product.download(
            progress_callback=progress_callback,
            output_dir=self.output_dir,
            extract=False,
        )

        # should be product id cast to str
        self.assertEqual(progress_callback.desc, "12345")

        # progress bar finished
        self.assertEqual(progress_callback.n, progress_callback.total)
        self.assertGreater(progress_callback.total, 0)

    def test_eoproduct_register_downloader(self):
        """eoproduct.register_donwloader must set download and auth plugins"""
        product = self._dummy_product()

        self.assertIsNone(product.downloader)
        self.assertIsNone(product.downloader_auth)

        downloader = mock.MagicMock()
        downloader_auth = mock.MagicMock()

        product.register_downloader(downloader, downloader_auth)

        self.assertEqual(product.downloader, downloader)
        self.assertEqual(product.downloader_auth, downloader_auth)

    def test_eoproduct_register_downloader_resolve_ok(self):
        """eoproduct.register_donwloader must resolve locations and properties"""
        downloadable_product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{
                        "downloadLink": "%(base_uri)s/is/resolved",
                        "otherProperty": "%(output_dir)s/also/resolved",
                    },
                )
            )
        )
        self.assertEqual(
            downloadable_product.location,
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.remote_location,
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.properties["downloadLink"],
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.properties["otherProperty"],
            f"{downloadable_product.downloader.config.output_dir}/also/resolved",
        )

    def test_eoproduct_register_downloader_resolve_ignored(self):
        """eoproduct.register_donwloader must ignore unresolvable locations and properties"""

        logger = logging.getLogger("eodag.product")
        with mock.patch.object(logger, "debug") as mock_debug:

            downloadable_product = self._dummy_downloadable_product(
                product=self._dummy_product(
                    properties=dict(
                        self.eoproduct_props,
                        **{
                            "downloadLink": "%(257B/cannot/be/resolved",
                            "otherProperty": "%(/%s/neither/resolved",
                        },
                    )
                )
            )
            self.assertEqual(downloadable_product.location, "%(257B/cannot/be/resolved")
            self.assertEqual(
                downloadable_product.remote_location, "%(257B/cannot/be/resolved"
            )
            self.assertEqual(
                downloadable_product.properties["downloadLink"],
                "%(257B/cannot/be/resolved",
            )
            self.assertEqual(
                downloadable_product.properties["otherProperty"],
                "%(/%s/neither/resolved",
            )

            needed_logs = [
                f"Could not resolve product.location ({downloadable_product.location})",
                f"Could not resolve product.remote_location ({downloadable_product.remote_location})",
                f"Could not resolve downloadLink property ({downloadable_product.properties['downloadLink']})",
                f"Could not resolve otherProperty property ({downloadable_product.properties['otherProperty']})",
            ]
            for needed_log in needed_logs:
                self.assertIn(needed_log, str(mock_debug.call_args_list))

    def test_eoproduct_repr_html(self):
        """eoproduct html repr must be correctly formatted"""
        product = self._dummy_product()
        product_repr = html.fromstring(product._repr_html_())
        self.assertIn("EOProduct", product_repr.xpath("//thead/tr/td")[0].text)

        # assets dict
        product.assets.update({"foo": {"href": "foo.href"}})
        assets_dict_repr = html.fromstring(product.assets._repr_html_())
        self.assertIn("AssetsDict", assets_dict_repr.xpath("//thead/tr/td")[0].text)

        # asset
        asset_repr = html.fromstring(product.assets._repr_html_())
        self.assertIn("Asset", asset_repr.xpath("//thead/tr/td")[0].text)

    def test_eoproduct_assets_get_values(self):
        """eoproduct.assets.get_values must return the expected values"""
        product = self._dummy_product()
        product.assets.update(
            {
                "foo": {"href": "foo.href"},
                "fooo": {"href": "fooo.href"},
                "foo?o,o": {"href": "foooo.href"},
            }
        )
        self.assertEqual(len(product.assets.get_values()), 3)
        self.assertEqual(len(product.assets.get_values("foo.*")), 3)
        self.assertEqual(len(product.assets.get_values("foo")), 1)
        self.assertEqual(product.assets.get_values("foo")[0]["href"], "foo.href")
        self.assertEqual(len(product.assets.get_values("foo?o,o")), 1)
        self.assertEqual(product.assets.get_values("foo?o,o")[0]["href"], "foooo.href")

    def test_eoproduct_none_properties(self):
        """eoproduct none properties must skipped"""
        product = self._dummy_product(
            properties={
                "geometry": "POINT (0 0)",
                "b_property": "b_value",
                "a_property": None,
            }
        )
        self.assertDictEqual(
            product.properties,
            {
                "b_property": "b_value",
            },
        )



# -*- coding: utf-8 -*-
# Copyright 2021, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from tests import TEST_RESOURCES_PATH, EODagTestCase
from tests.context import EOProduct, NoDriver, Sentinel1Driver


class TestEOProductDriverSentinel1Driver(EODagTestCase):
    def setUp(self):
        super(TestEOProductDriverSentinel1Driver, self).setUp()
        self.product = EOProduct(
            self.provider, self.eoproduct_props, productType="S1_SAR_OCN"
        )
        self.product.properties["title"] = os.path.join(
            TEST_RESOURCES_PATH,
            "products",
            "S2A_MSIL1C_20180101T105441_N0206_R051_T31TDH_20180101T124911",
        )

    def test_driver_s1_init(self):
        """The appropriate driver must have been set"""
        self.assertIsInstance(self.product.driver, Sentinel1Driver)
        self.assertTrue(hasattr(self.product.driver, "legacy"))
        try:
            # import from eodag-cube if installed
            from eodag_cube.api.product.drivers.base import (  # pyright: ignore[reportMissingImports]; isort: skip
                DatasetDriver as DatasetDriver_cube,
            )

            self.assertIsInstance(self.product.driver.legacy, DatasetDriver_cube)
        except ImportError:
            self.assertIsInstance(self.product.driver.legacy, NoDriver)

    def test_driver_s1_guess_asset_key_and_roles(self):
        """The driver must guess appropriate asset key and roles"""
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles("", self.product),
            (None, None),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "2018/1/28/0/ew-hv.tiff", self.product
            ),
            ("HV", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "http://foo/1/28/0/iw-hh.tif", self.product
            ),
            ("HH", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/s1a-vv.tiff", self.product
            ),
            ("VV", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/rfi-ew-hh.xml", self.product
            ),
            ("rfi-hh.xml", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/thumbnail.png", self.product
            ),
            ("thumbnail.png", ["thumbnail"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/quick-look.jpg", self.product
            ),
            ("quick-look.jpg", ["overview"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/foo.bar", self.product
            ),
            ("foo.bar", ["auxiliary"]),
        )



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import io
import logging
import os
import uuid
from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from zipfile import ZIP_STORED, ZipFile

import botocore
import botocore.exceptions
from concurrent.futures import FIRST_COMPLETED, ThreadPoolExecutor, wait
from zipstream import ZipStream

from eodag.plugins.authentication.aws_auth import AwsAuth
from eodag.utils import (
    StreamResponse,
    get_bucket_name_and_prefix,
    guess_file_type,
    parse_le_uint16,
    parse_le_uint32,
)
from eodag.utils.exceptions import (
    AuthenticationError,
    InvalidDataError,
    NotAvailableError,
)

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Literal, Optional
    from zipfile import ZipInfo

    from mypy_boto3_s3.client import S3Client

    from eodag.api.product import EOProduct  # type: ignore

logger = logging.getLogger("eodag.utils.s3")

MIME_OCTET_STREAM = "application/octet-stream"


def fetch_range(
    bucket_name: str, key_name: str, start: int, end: int, client_s3: S3Client
) -> bytes:
    """
    Range-fetches a S3 key.

    :param bucket_name: Bucket name of the object to fetch
    :param key_name: Key name of the object to fetch
    :param start: Start byte position to fetch
    :param end: End byte position to fetch
    :param client_s3: s3 client used to fetch the object
    :returns: Object bytes
    """
    response = client_s3.get_object(
        Bucket=bucket_name, Key=key_name, Range="bytes=%d-%d" % (start, end)
    )
    return response["Body"].read()


@dataclass
class S3FileInfo:
    """
    Describe a S3 object with basic f_info and its download state.
    """

    size: int
    key: str
    bucket_name: str
    #: Path inside the ZIP archive if the file is stored inside a ZIP.
    zip_filepath: Optional[str] = None
    #: Offset in the ZIP archive where the file data starts.
    data_start_offset: int = 0
    #: MIME type of the file, defaulting to application/octet-stream.
    #: It can be updated based on the file extension or content type.
    data_type: str = MIME_OCTET_STREAM
    #: Relative path of the file, if applicable (e.g., inside a ZIP archive).
    rel_path: Optional[str] = None

    # These fields hold the state for downloading
    #: Offset in the logical (global) file stream where this file starts.
    file_start_offset: int = 0
    #: Mapping of futures to their start byte offsets, used to track download progress.
    #: Each future corresponds to a chunk of data being downloaded.
    #: The key is the future object, and the value is the start byte offset of that
    #: chunk in the logical file stream.
    futures: dict = field(default_factory=dict)
    #: Buffers for downloaded data chunks, mapping start byte offsets to the actual data.
    #: This allows for partial downloads and efficient memory usage.
    #: The key is the start byte offset, and the value is the bytes data for that
    #: offset. This is used to yield data in the correct order during streaming.
    #: It is updated as chunks are downloaded.
    buffers: dict[int, bytes] = field(default_factory=dict)
    #: The next offset to yield in the file, used to track progress during downloading
    #: and yielding chunks. It starts at 0 and is updated as data is yielded.
    #: This allows the streaming process to continue from where it left off,
    #: ensuring that all data is eventually yielded without duplication.
    next_yield: int = 0


def _prepare_file_in_zip(f_info: S3FileInfo, s3_client: S3Client):
    """Update file information with the offset and size of the file inside the zip archive"""

    splitted_path = f_info.key.split(".zip!")
    f_info.key = f"{splitted_path[0]}.zip"
    f_info.zip_filepath = splitted_path[-1]  # file path inside the ZIP archive

    f_info.data_start_offset, f_info.size = file_position_from_s3_zip(
        f_info.bucket_name,
        f_info.key,
        s3_client,
        f_info.zip_filepath,
    )


def _compute_file_ranges(
    file_info: S3FileInfo,
    byte_range: tuple[Optional[int], Optional[int]],
    range_size: int,
) -> Optional[list[tuple[int, int]]]:
    """
    Compute the byte ranges to download for a single file, considering the overall requested range.

    This function calculates which byte ranges within the given file should be downloaded,
    based on the global requested byte range (`byte_range`) and the size of each chunk (`range_size`).
    It accounts for possible offsets if the file is part of a ZIP archive or not aligned at offset zero.

    :param file_info: The S3FileInfo object containing file metadata, including size, data offset,
                      and its starting offset in the full logical file stream.
    :param byte_range: A tuple (start, end) specifying the requested global byte range, where either
                       value may be None to indicate an open-ended range.
    :param range_size: The size of each download chunk in bytes.
    :returns: A list of (start, end) tuples indicating byte ranges to download within this file,
              or None if the file lies completely outside the requested range.
    """
    file_size = file_info.size
    file_start_offset = file_info.file_start_offset
    file_end_offset = file_start_offset + file_size - 1

    range_start, range_end = byte_range

    # Check if the file overlaps with the requested range
    if range_end is not None and range_start is not None:
        if file_end_offset < range_start or file_start_offset > range_end:
            return None  # No overlap, skip this file

    start = 0
    end = file_size - 1

    # Adjust start and end based on the requested range
    if range_start is not None:
        start = max(0, range_start - file_start_offset)
    if range_end is not None:
        end = min(file_size - 1, range_end - file_start_offset)

    start += file_info.data_start_offset
    end += file_info.data_start_offset

    # Compute the ranges in chunks
    ranges = []
    for chunk_start in range(start, end + 1, range_size):
        chunk_end = min(chunk_start + range_size - 1, end)
        ranges.append((chunk_start, chunk_end))

    return ranges


def _chunks_from_s3_objects(
    s3_client: S3Client,
    files_info: list[S3FileInfo],
    byte_range: tuple[Optional[int], Optional[int]],
    range_size: int,
    executor: ThreadPoolExecutor,
) -> Iterator[tuple[int, Iterator[bytes]]]:
    """Download chunks from S3 objects in parallel, respecting byte ranges and file order."""
    # Prepare ranges and futures per file
    for f_info in files_info:
        ranges = _compute_file_ranges(f_info, byte_range, range_size)

        if not ranges:
            # Mark as inactive (no futures)
            f_info.futures = {}
            f_info.buffers = {}
            f_info.next_yield = 0
            continue

        f_info.buffers = {}
        f_info.next_yield = 0

        futures = {}
        # start,end are absolute offsets in the S3 object (data_start_offset already applied)
        for start, end in ranges:
            future = executor.submit(
                fetch_range,
                f_info.bucket_name,
                f_info.key,
                start,
                end,
                s3_client,
            )
            # Track both start and end so we can compute the yielded length precisely
            futures[future] = (start, end)

        f_info.futures = futures

    # Keep only files that actually have something to download
    active_indices = [i for i, fi in enumerate(files_info) if fi.futures]

    # Combine all futures to wait on globally
    all_futures = {
        fut: (f_info, start, end)
        for f_info in (files_info[i] for i in active_indices)
        for fut, (start, end) in f_info.futures.items()
    }

    def make_chunks_generator(target_info: S3FileInfo) -> Iterator[bytes]:
        """Create a generator bound to a specific file info (no late-binding bug)."""
        info = target_info  # bind
        nonlocal all_futures
        while info.next_yield < info.size:
            # First, try to flush anything already buffered for this file
            next_start = info.next_yield
            flushed = False
            while next_start in info.buffers:
                chunk = info.buffers.pop(next_start)
                if not isinstance(chunk, bytes):
                    raise InvalidDataError(
                        f"Expected bytes, got {type(chunk).__name__} in stream chunks: {chunk}"
                    )
                yield chunk
                next_start += len(chunk)
                info.next_yield = next_start
                flushed = True

            if info.next_yield >= info.size:
                break

            # If we flushed something, loop back to try again before waiting
            if flushed:
                continue

            # Nothing to flush for this file: wait for more futures to complete globally
            if not all_futures:
                # No more incoming data anywhere; stop to avoid waiting on an empty set
                break

            done, _ = wait(all_futures.keys(), return_when=FIRST_COMPLETED)
            for fut in done:
                f_info, start, end = all_futures.pop(fut)
                data = fut.result()
                # Store buffer with a key relative to the start of the file data
                rel_start = start - f_info.data_start_offset
                f_info.buffers[rel_start] = data

    # Yield per-file generators with their original indices
    for idx in active_indices:
        yield idx, make_chunks_generator(files_info[idx])


def _build_stream_response(
    zip_filename: str,
    files_info: list[S3FileInfo],
    files_iterator: Iterator[tuple[int, Iterator[bytes]]],
    compress: Literal["zip", "raw", "auto"],
    executor: ThreadPoolExecutor,
) -> StreamResponse:
    """
    Build a streaming HTTP response for one or multiple files from S3, supporting ZIP, raw, and multipart formats.

    The response format depends on the `compress` parameter and the number of files:

    - If `compress` is "zip" or "auto" with multiple files, returns a ZIP archive containing all files.
    - If `compress` is "raw" and multiple files, returns a multipart/mixed response with each file as a part.
    - If only one file is present and `compress` is "raw" or "auto", streams the file directly with its MIME type.

    Response formats:

    - ZIP archive (Content-Type: application/zip) with Content-Disposition for download.
    - Multipart/mixed (Content-Type: multipart/mixed; boundary=...) with each file as a part.
    - Single raw file stream with its MIME type and Content-Disposition for download.

    :param zip_filename: Base filename to use for the ZIP archive (without extension).
    :param files_info: List of S3FileInfo objects describing each file (metadata, MIME type, etc.).
    :param files_iterator: Iterator yielding (file_index, chunk_iterator) for streaming file contents.
    :param compress: Output format:
        - "zip": Always produce a ZIP archive.
        - "raw": Stream files directly, as a single file or multipart.
        - "auto": ZIP if multiple files, raw if single file.
    :param executor: Executor used for concurrent streaming and cleanup.
    :return: Streaming HTTP response with appropriate content, headers, and media type.
    """

    def _wrap_generator_with_cleanup(
        generator: Iterable[bytes], executor: ThreadPoolExecutor
    ) -> Iterator[bytes]:
        try:
            yield from generator
        finally:
            executor.shutdown(wait=True)

    def _build_response(
        content_gen: Iterable[bytes],
        media_type: str,
        filename: Optional[str] = None,
        size: Optional[int] = None,
    ) -> StreamResponse:
        return StreamResponse(
            content=_wrap_generator_with_cleanup(content_gen, executor),
            media_type=media_type,
            headers={"Accept-Ranges": "bytes"},
            filename=filename,
            size=size,
        )

    zip_response = (len(files_info) > 1 and compress == "auto") or compress == "zip"

    if zip_response:
        zs = ZipStream(sized=True)
        for index, chunks_generator in files_iterator:
            file_info = files_info[index]
            file_path = file_info.rel_path or file_info.key
            zs.add(chunks_generator, file_path, size=file_info.size)

        return _build_response(
            content_gen=zs,
            media_type="application/zip",
            filename=f"{zip_filename}.zip",
            size=len(zs),
        )
    elif len(files_info) > 1:
        boundary = uuid.uuid4().hex

        def multipart_stream():
            current_index = -1
            for index, chunks_generator in files_iterator:
                if index != current_index:
                    if current_index != -1:
                        yield b"\r\n"
                    filename = os.path.basename(files_info[index].key)
                    yield (
                        f"--{boundary}\r\n"
                        f'Content-Disposition: attachment; filename="{filename}"\r\n'
                        f"Content-Type: {files_info[index].data_type}\r\n\r\n"
                    ).encode()
                    current_index = index
                yield from chunks_generator
            yield f"\r\n--{boundary}--\r\n".encode()

        return _build_response(
            content_gen=multipart_stream(),
            media_type=f"multipart/mixed; boundary={boundary}",
        )
    else:
        index, chunks_generator = next(files_iterator)
        first_chunk = next(chunks_generator)
        filename = os.path.basename(files_info[index].key)

        def single_file_stream() -> Iterator[bytes]:
            yield first_chunk
            yield from chunks_generator

        return _build_response(
            content_gen=single_file_stream(),
            media_type=files_info[index].data_type,
            filename=filename,
        )


def stream_download_from_s3(
    s3_client: S3Client,
    files_info: list[S3FileInfo],
    byte_range: tuple[Optional[int], Optional[int]] = (None, None),
    compress: Literal["zip", "raw", "auto"] = "auto",
    zip_filename: str = "archive",
    range_size: int = 1024**2 * 8,
    max_workers: int = 8,
) -> StreamResponse:
    """
    Stream data from one or more S3 objects in chunks, with support for global byte ranges.

    This function provides efficient streaming download of S3 objects with support for:

    * Single file streaming with direct MIME type detection
    * Multiple file streaming as ZIP archives
    * Byte range requests for partial content
    * Files within ZIP archives (using ``.zip!`` notation)
    * Concurrent chunk downloading for improved performance
    * Memory-efficient streaming without loading entire files

    The response format depends on the compress parameter and number of files:

    * Single file + ``compress="raw"`` or ``"auto"``: streams file directly with detected MIME type
    * Multiple files + ``compress="zip"`` or ``"auto"``: creates ZIP archive containing all files
    * ``compress="zip"``: always creates ZIP archive regardless of file count

    For files stored within ZIP archives, use the ``.zip!`` notation in the ``S3FileInfo.key``:
    ``"path/to/archive.zip!internal/file.txt"``

    :param s3_client: Boto3 S3 client instance for making requests
    :param files_info: List of S3FileInfo objects describing files to download.
        Each object must contain at minimum: ``bucket_name``, ``key``, and ``size``.
        Optional fields include: ``data_type``, ``rel_path``, ``zip_filepath``.
    :param byte_range: Global byte range to download as ``(start, end)`` tuple.
        ``None`` values indicate open-ended ranges.
        Applied across the logical concatenation of all files.
    :param compress: Output format control:

        * ``"zip"``: Always create ZIP archive
        * ``"raw"``: Stream files directly (single) or as multipart (multiple)
        * ``"auto"``: ZIP for multiple files, raw for single file

    :param zip_filename: Base filename for ZIP archives (without ``.zip`` extension).
        Only used when creating ZIP archives.
    :param range_size: Size of each download chunk in bytes. Larger chunks reduce
        request overhead but use more memory. Default: 8MB.
    :param max_workers: Maximum number of concurrent download threads.
        Higher values improve throughput for multiple ranges.
    :return: StreamResponse object containing:

        * ``content``: Iterator of bytes for the streaming response
        * ``media_type``: MIME type (``"application/zip"`` for archives, detected type for single files)
        * ``headers``: HTTP headers including Content-Disposition for downloads

    :rtype: StreamResponse
    :raises InvalidDataError: If ZIP file structures are malformed
    :raises NotAvailableError: If S3 objects cannot be accessed
    :raises AuthenticationError: If S3 credentials are invalid
    :raises NotImplementedError: If compressed files within ZIP archives are encountered

    Example usage:

    .. code-block:: python

        import boto3
        from eodag.utils.s3 import stream_download_from_s3, S3FileInfo

        # Create S3 client
        s3_client = boto3.client('s3')

        # Single file download
        files = [S3FileInfo(bucket_name="bucket", key="file.txt", size=1024)]
        response = stream_download_from_s3(s3_client, files)

        # Multiple files as ZIP archive
        files = [
            S3FileInfo(bucket_name="bucket", key="file1.txt", size=1024),
            S3FileInfo(bucket_name="bucket", key="file2.txt", size=2048)
        ]
        response = stream_download_from_s3(s3_client, files, compress="zip")

        # File within ZIP archive
        files = [S3FileInfo(
            bucket_name="bucket",
            key="archive.zip!internal.txt",
            size=512
        )]
        response = stream_download_from_s3(s3_client, files)

        # Process streaming response
        for chunk in response.content:
            # Handle chunk data
            pass
    """

    executor = ThreadPoolExecutor(max_workers=max_workers)

    # Prepare all files
    offset = 0
    for f_info in files_info:
        if ".zip!" in f_info.key:
            _prepare_file_in_zip(f_info, s3_client)

        f_info.file_start_offset = offset
        offset += f_info.size

        if not f_info.data_type or f_info.data_type == MIME_OCTET_STREAM:
            guessed = guess_file_type(f_info.key)
            f_info.data_type = guessed or MIME_OCTET_STREAM

    # Create the files iterator using the original approach
    files_iterator = _chunks_from_s3_objects(
        s3_client, files_info, byte_range, range_size, executor
    )

    # Use the existing _build_stream_response function with the additional parameters
    return _build_stream_response(
        zip_filename=zip_filename,
        files_info=files_info,
        files_iterator=files_iterator,
        compress=compress,
        executor=executor,
    )


def update_assets_from_s3(
    product: EOProduct,
    auth: AwsAuth,
    s3_endpoint: Optional[str] = None,
    content_url: Optional[str] = None,
) -> None:
    """Update ``EOProduct.assets`` using content listed in its ``remote_location`` or given
    ``content_url``.

    If url points to a zipped archive, its content will also be be listed.

    :param product: product to update
    :param auth: Authentication plugin
    :param s3_endpoint: s3 endpoint if not hosted on AWS
    :param content_url: s3 URL pointing to the content that must be listed (defaults to
                        ``product.remote_location`` if empty)
    """

    if content_url is None:
        content_url = product.remote_location

    bucket, prefix = get_bucket_name_and_prefix(content_url)

    if bucket is None or prefix is None:
        logger.debug(f"No s3 prefix could guessed from {content_url}")
        return None

    try:

        logger.debug("Listing assets in %s", prefix)
        s3_client = auth.get_s3_client()

        if prefix.endswith(".zip"):
            # List prefix zip content
            assets_urls = [
                f"zip+s3://{bucket}/{prefix}!{f.filename}"
                for f in list_files_in_s3_zipped_object(bucket, prefix, s3_client)
            ]
        else:
            # List files in prefix
            assets_urls = [
                f"s3://{bucket}/{obj['Key']}"
                for obj in s3_client.list_objects(
                    Bucket=bucket, Prefix=prefix, MaxKeys=300
                ).get("Contents", [])
            ]

        for asset_url in assets_urls:
            out_of_zip_url = asset_url.split("!")[-1]
            key, roles = product.driver.guess_asset_key_and_roles(
                out_of_zip_url, product
            )

            if key and key not in product.assets:
                product.assets[key] = {
                    "title": key,  # Normalize title with key
                    "roles": roles,
                    "href": asset_url,
                }
                if mime_type := guess_file_type(asset_url):
                    product.assets[key]["type"] = mime_type

        # sort assets
        product.assets.data = dict(sorted(product.assets.data.items()))

        # update driver
        product.driver = product.get_driver()

    except botocore.exceptions.ClientError as e:
        if hasattr(auth.config, "auth_error_code") and str(
            auth.config.auth_error_code
        ) in str(e):
            raise AuthenticationError(
                f"Authentication failed on {s3_endpoint} s3"
            ) from e
        raise NotAvailableError(
            f"assets for product {prefix} could not be found"
        ) from e


# ----- ZIP section -----


def open_s3_zipped_object(
    bucket_name: str,
    key_name: str,
    s3_client,
    zip_size: Optional[int] = None,
    partial: bool = True,
) -> tuple[ZipFile, bytes]:
    """
    Fetches the central directory and EOCD (End Of Central Directory) from an S3 object and opens a ZipFile in memory.

    This function retrieves the ZIP file's central directory and EOCD by performing range requests on the S3 object.
    It supports partial fetching (only the central directory and EOCD) for efficiency, or full ZIP download if needed.

    :param bucket_name: Name of the S3 bucket containing the ZIP file.
    :param key_name: Key (path) of the ZIP file in the S3 bucket.
    :param s3_client: S3 client instance used to perform range requests.
    :param zip_size: Size of the ZIP file in bytes. If None, it will be determined via a HEAD request.
    :param partial: If True, only fetch the central directory and EOCD. If False, fetch the entire ZIP file.
    :return: Tuple containing the opened ZipFile object and the central directory bytes.
    :raises InvalidDataError: If the EOCD signature is not found in the last 64KB of the file.
    """
    # EOCD is at least 22 bytes, but can be longer if ZIP comment exists.
    # For simplicity, we fetch last 64KB max (max EOCD + comment length allowed by ZIP spec)
    if zip_size is None:
        response = s3_client.head_object(Bucket=bucket_name, Key=key_name)
        zip_size = int(response["ContentLength"])

    fetch_size = min(65536, zip_size)
    eocd_search = fetch_range(
        bucket_name, key_name, zip_size - fetch_size, zip_size - 1, s3_client
    )

    # Find EOCD signature from end: 0x06054b50 (little endian)
    eocd_signature = b"\x50\x4b\x05\x06"
    eocd_offset = eocd_search.rfind(eocd_signature)
    if eocd_offset == -1:
        raise InvalidDataError("EOCD signature not found in last 64KB of the file.")

    eocd = eocd_search[eocd_offset : eocd_offset + 22]

    cd_size = parse_le_uint32(eocd[12:16])
    cd_start = parse_le_uint32(eocd[16:20])

    # Fetch central directory
    cd_data = fetch_range(
        bucket_name, key_name, cd_start, cd_start + cd_size - 1, s3_client
    )

    zip_data = (
        cd_data + eocd
        if partial
        else fetch_range(bucket_name, key_name, 0, zip_size - 1, s3_client)
    )
    zipf = ZipFile(io.BytesIO(zip_data))
    return zipf, cd_data


def _parse_central_directory_entry(cd_data: bytes, offset: int) -> dict[str, int]:
    """
    Parse one central directory file header entry starting at offset.
    Returns dict with relative local header offset and sizes.
    """
    signature = cd_data[offset : offset + 4]
    if signature != b"PK\x01\x02":
        raise InvalidDataError("Bad central directory file header signature")

    filename_len = parse_le_uint16(cd_data[offset + 28 : offset + 30])
    extra_len = parse_le_uint16(cd_data[offset + 30 : offset + 32])
    comment_len = parse_le_uint16(cd_data[offset + 32 : offset + 34])

    relative_offset = parse_le_uint32(cd_data[offset + 42 : offset + 46])

    header_size = 46 + filename_len + extra_len + comment_len

    return {
        "relative_offset": relative_offset,
        "header_size": header_size,
        "filename_len": filename_len,
        "extra_len": extra_len,
        "comment_len": comment_len,
        "total_size": header_size,
    }


def _parse_local_file_header(local_header_bytes: bytes) -> int:
    """
    Parse local file header to find total header size:
    fixed 30 bytes + filename length + extra field length
    """
    if local_header_bytes[0:4] != b"PK\x03\x04":
        raise InvalidDataError("Bad local file header signature")

    filename_len = parse_le_uint16(local_header_bytes[26:28])
    extra_len = parse_le_uint16(local_header_bytes[28:30])
    total_size = 30 + filename_len + extra_len
    return total_size


def file_position_from_s3_zip(
    s3_bucket: str,
    object_key: str,
    s3_client,
    target_filepath: str,
) -> tuple[int, int]:
    """
    Get the start position and size of a specific file inside a ZIP archive stored in S3.
    This function assumes the file is uncompressed (ZIP_STORED).

    The returned tuple contains:

    - **file_data_start**: The byte offset where the file data starts in the ZIP archive.
    - **file_size**: The size of the file in bytes.

    :param s3_bucket: The S3 bucket name.
    :param object_key: The S3 object key for the ZIP file.
    :param s3_client: The Boto3 S3 client.
    :param target_filepath: The file path inside the ZIP archive to locate.
    :return: A tuple (file_data_start, file_size)
    :raises FileNotFoundError: If the target file is not found in the ZIP archive.
    :raises NotImplementedError: If the file is not uncompressed (ZIP_STORED)
    """
    zipf, cd_data = open_s3_zipped_object(s3_bucket, object_key, s3_client)

    # Find file in zipf.filelist to get its index and f_info
    target_info = None
    cd_offset = 0
    for fi in zipf.filelist:
        if fi.filename == target_filepath:
            target_info = fi
            break
        # 46 is the fixed size (in bytes) of the Central Directory File Header according to the ZIP spec
        cd_entry_len = (
            46 + len(fi.filename.encode("utf-8")) + len(fi.extra) + len(fi.comment)
        )
        cd_offset += cd_entry_len

    zipf.close()

    if target_info is None:
        raise FileNotFoundError(f"File {target_filepath} not found in ZIP archive")

    if target_info.compress_type != ZIP_STORED:
        raise NotImplementedError("Only uncompressed files (ZIP_STORED) are supported.")

    # Parse central directory entry to get relative local header offset
    cd_entry = cd_data[
        cd_offset : cd_offset
        + (
            46
            + len(target_info.filename.encode("utf-8"))
            + len(target_info.extra)
            + len(target_info.comment)
        )
    ]
    cd_entry_info = _parse_central_directory_entry(cd_entry, 0)

    local_header_offset = cd_entry_info["relative_offset"]

    # Fetch local file header from S3 (at least 30 bytes + filename + extra field)
    # We'll fetch 4 KB max to cover large filenames/extra fields safely
    local_header_fetch_size = 4096
    local_header_bytes = fetch_range(
        s3_bucket,
        object_key,
        local_header_offset,
        local_header_offset + local_header_fetch_size - 1,
        s3_client,
    )

    local_header_size = _parse_local_file_header(local_header_bytes)

    # Calculate file data start and end offsets
    file_data_start = local_header_offset + local_header_size

    return file_data_start, target_info.file_size


def list_files_in_s3_zipped_object(
    bucket_name: str, key_name: str, s3_client: S3Client
) -> list[ZipInfo]:
    """
    List files in s3 zipped object, without downloading it.

    See https://stackoverflow.com/questions/41789176/how-to-count-files-inside-zip-in-aws-s3-without-downloading-it;
    Based on https://stackoverflow.com/questions/51351000/read-zip-files-from-s3-without-downloading-the-entire-file

    :param bucket_name: Bucket name of the object to fetch
    :param key_name: Key name of the object to fetch
    :param s3_resource: s3 resource used to fetch the object
    :returns: List of files in zip
    """
    zip_file, _ = open_s3_zipped_object(bucket_name, key_name, s3_client)
    with zip_file:
        logger.debug("Found %s files in %s" % (len(zip_file.filelist), key_name))
        return zip_file.filelist



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from shapely.errors import ShapelyError

from eodag.plugins.crunch.base import Crunch
from eodag.utils import get_geometry_from_various

if TYPE_CHECKING:
    from eodag.api.product import EOProduct

logger = logging.getLogger("eodag.crunch.overlap")


class FilterOverlap(Crunch):
    """FilterOverlap cruncher

    Filter products, retaining only those that are overlapping with the search_extent

    :param config: Crunch configuration may contain the following parameters which are mutually exclusive:

        * ``minimum_overlap`` (``Union[float, str]``): minimal overlap percentage; default: ``"0"``
        * ``contains`` (``bool``): ``True`` if product geometry contains the search area; default: ``False``
        * ``intersects`` (``bool``): ``True`` if product geometry intersects the search area; default: ``False``
        * ``within`` (``bool``): ``True`` if product geometry is within the search area; default: ``False``
    """

    def proceed(
        self, products: list[EOProduct], **search_params: Any
    ) -> list[EOProduct]:
        """Execute crunch: Filter products, retaining only those that are overlapping with the search_extent

        :param products: A list of products resulting from a search
        :param search_params: Search criteria that must contain `geometry`
        :returns: The filtered products
        """
        logger.debug("Start filtering for overlapping products")
        filtered: list[EOProduct] = []
        add_to_filtered = filtered.append

        search_geom = get_geometry_from_various(**search_params)
        if not search_geom:
            logger.warning(
                "geometry not found in cruncher arguments, filtering disabled."
            )
            return products
        minimum_overlap = float(self.config.__dict__.get("minimum_overlap", "0"))
        contains = self.config.__dict__.get("contains", False)
        intersects = self.config.__dict__.get("intersects", False)
        within = self.config.__dict__.get("within", False)

        if contains and (within or intersects) or (within and intersects):
            logger.warning(
                "contains, intersects and within parameters are mutually exclusive"
            )
            return products
        elif (
            minimum_overlap > 0
            and minimum_overlap < 100
            and (contains or within or intersects)
        ):
            logger.warning(
                "minimum_overlap will be ignored because of contains/intersects/within usage"
            )
        elif not contains and not within and not intersects:
            logger.debug("Minimum overlap is: {} %".format(minimum_overlap))

        logger.debug("Initial requested extent area: %s", search_geom.area)
        if search_geom.area == 0:
            logger.debug(
                "No product can overlap a requested extent that is not a polygon (i.e with area=0)"
            )
        else:
            for product in products:
                logger.debug("Uncovered extent area: %s", search_geom.area)
                if product.search_intersection:
                    intersection = product.search_intersection
                    product_geometry = product.geometry
                else:  # Product geometry may be invalid
                    if not product.geometry.is_valid:
                        logger.debug(
                            "Trying our best to deal with invalid geometry on product: %r",
                            product,
                        )
                        product_geometry = product.geometry.buffer(0)
                        try:
                            intersection = search_geom.intersection(product_geometry)
                        except ShapelyError:
                            logger.debug(
                                "Product geometry still invalid. Overlap test restricted to containment"
                            )
                            if search_geom.contains(product_geometry):
                                logger.debug(
                                    "Product %r overlaps the search extent. Adding it to filtered results"
                                )
                                add_to_filtered(product)
                            continue
                    else:
                        product_geometry = product.geometry
                        intersection = search_geom.intersection(product_geometry)

                if (
                    (contains and product_geometry.contains(search_geom))
                    or (within and product_geometry.within(search_geom))
                    or (intersects and product_geometry.intersects(search_geom))
                ):
                    add_to_filtered(product)
                    continue
                elif contains or within or intersects:
                    continue

                ipos = (intersection.area / search_geom.area) * 100
                ipop = (intersection.area / product_geometry.area) * 100
                logger.debug(
                    "Intersection of product extent and search extent covers %f percent of the search extent "
                    "area",
                    ipos,
                )
                logger.debug(
                    "Intersection of product extent and search extent covers %f percent of the product extent "
                    "area",
                    ipop,
                )
                if any(
                    (
                        search_geom.contains(product.geometry),
                        ipos >= minimum_overlap,
                        ipop >= minimum_overlap,
                    )
                ):
                    logger.debug(
                        "Product %r overlaps the search extent by the specified constraint. Adding it to "
                        "filtered results",
                        product,
                    )
                    add_to_filtered(product)
                else:
                    logger.debug(
                        "Product %r does not overlaps the search extent by the specified constraint. "
                        "Skipping it",
                        product,
                    )
        logger.info("Finished filtering products. %s resulting products", len(filtered))
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2023, CS Systemes d'Information, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EODAG REST utils"""
from __future__ import annotations

import glob
import logging
import os
from io import BufferedReader
from shutil import make_archive, rmtree
from typing import TYPE_CHECKING, Any, Callable, Iterator, NamedTuple, Optional, Union
from urllib.parse import unquote_plus, urlencode

import orjson
from fastapi import Request
from pydantic import ValidationError as pydanticValidationError

from eodag.plugins.crunch.filter_latest_intersect import FilterLatestIntersect
from eodag.plugins.crunch.filter_latest_tpl_name import FilterLatestByName
from eodag.plugins.crunch.filter_overlap import FilterOverlap
from eodag.utils import StreamResponse
from eodag.utils.dates import get_date, get_datetime
from eodag.utils.exceptions import ValidationError

if TYPE_CHECKING:
    from eodag.rest.types.stac_search import SearchPostRequest

# exportable content
__all__ = ["get_date", "get_datetime"]

logger = logging.getLogger("eodag.rest.utils")

# Path of the liveness endpoint
LIVENESS_PROBE_PATH = "/_mgmt/ping"


class Cruncher(NamedTuple):
    """Type hinted Cruncher namedTuple"""

    clazz: Callable[..., Any]
    config_params: list[str]


crunchers = {
    "latestIntersect": Cruncher(FilterLatestIntersect, []),
    "latestByName": Cruncher(FilterLatestByName, ["name_pattern"]),
    "overlap": Cruncher(FilterOverlap, ["minimum_overlap"]),
}


def format_pydantic_error(e: pydanticValidationError) -> str:
    """Format Pydantic ValidationError

    :param e: A Pydantic ValidationError object
    :tyype e: pydanticValidationError
    """
    error_header = f"{e.error_count()} error(s). "

    error_messages = [
        f'{err["loc"][0]}: {err["msg"]}' if err["loc"] else err["msg"]
        for err in e.errors()
    ]
    return error_header + "; ".join(set(error_messages))


def is_dict_str_any(var: Any) -> bool:
    """Verify whether the variable is of type dict[str, Any]"""
    if isinstance(var, dict):
        return all(isinstance(k, str) for k in var.keys())  # type: ignore
    return False


def str2list(v: Optional[str]) -> Optional[list[str]]:
    """Convert string to list base on , delimiter."""
    if v:
        return v.split(",")
    return None


def str2json(k: str, v: Optional[str] = None) -> Optional[dict[str, Any]]:
    """decoding a URL parameter and then parsing it as JSON."""
    if not v:
        return None
    try:
        return orjson.loads(unquote_plus(v))
    except orjson.JSONDecodeError as e:
        raise ValidationError(f"{k}: Incorrect JSON object") from e


def flatten_list(nested_list: Union[Any, list[Any]]) -> list[Any]:
    """Flatten a nested list structure into a single list."""
    if not isinstance(nested_list, list):
        return [nested_list]
    else:
        flattened: list[Any] = []
        for element in nested_list:
            flattened.extend(flatten_list(element))
        return flattened


def list_to_str_list(input_list: list[Any]) -> list[str]:
    """Attempt to convert a list of any type to a list of strings."""
    try:
        # Try to convert each element to a string
        return [str(element) for element in input_list]
    except Exception as e:
        # Raise an exception if any element cannot be converted
        raise TypeError(f"Failed to convert to list[str]: {e}") from e


def get_next_link(
    request: Request,
    search_request: SearchPostRequest,
    total_results: Optional[int],
    items_per_page: int,
) -> Optional[dict[str, Any]]:
    """Generate next link URL and body"""
    body = search_request.model_dump(exclude_none=True)
    if "bbox" in body:
        # bbox is tuple
        body["bbox"] = list(body["bbox"])

    params = dict(request.query_params)

    page = int(body.get("page", 0) or params.get("page", 0)) or 1

    if total_results is None or items_per_page * page >= total_results:
        return None

    url = str(request.state.url)
    if request.method == "POST":
        body["page"] = page + 1
    else:
        params["page"] = str(page + 1)
        url += f"?{urlencode(params)}"

    next: dict[str, Any] = {
        "rel": "next",
        "href": url,
        "title": "Next page",
        "method": request.method,
        "type": "application/geo+json",
    }
    if request.method == "POST":
        next["body"] = body
    return next


def read_file_chunks_and_delete(
    opened_file: BufferedReader, chunk_size: int = 64 * 1024
) -> Iterator[bytes]:
    """Yield file chunks and delete file when finished."""
    while True:
        data = opened_file.read(chunk_size)
        if not data:
            opened_file.close()
            os.remove(opened_file.name)
            logger.debug("%s deleted after streaming complete", opened_file.name)
            break
        yield data
    yield data


def file_to_stream(
    file_path: str,
) -> StreamResponse:
    """Break a file into chunck and return it as a byte stream"""
    if os.path.isdir(file_path):
        # do not zip if dir contains only one file
        all_filenames = [
            f
            for f in glob.glob(os.path.join(file_path, "**", "*"), recursive=True)
            if os.path.isfile(f)
        ]
        if len(all_filenames) == 1:
            filepath_to_stream = all_filenames[0]
        else:
            filepath_to_stream = f"{file_path}.zip"
            logger.debug(
                "Building archive for downloaded product path %s",
                filepath_to_stream,
            )
            make_archive(file_path, "zip", file_path)
            rmtree(file_path)
    else:
        filepath_to_stream = file_path

    filename = os.path.basename(filepath_to_stream)
    return StreamResponse(
        content=read_file_chunks_and_delete(open(filepath_to_stream, "rb")),
        filename=filename,
    )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import json
import os
import unittest
from tempfile import TemporaryDirectory
from unittest.mock import Mock

import pytest

from eodag.rest.types.stac_search import SearchPostRequest
from eodag.utils.exceptions import ValidationError
from tests import TEST_RESOURCES_PATH, mock
from tests.context import SearchResult
from tests.utils import mock_request


class TestStacCore(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestStacCore, cls).setUpClass()

        # Mock home and eodag conf directory to tmp dir
        cls.tmp_home_dir = TemporaryDirectory()
        cls.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, return_value=cls.tmp_home_dir.name
        )
        cls.expanduser_mock.start()

        # import after having mocked home_dir because it launches http server (and EODataAccessGateway)
        import eodag.rest.core as rest_core

        importlib.reload(rest_core)

        cls.rest_core = rest_core

        search_results_file = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result_peps.geojson"
        )
        with open(search_results_file, encoding="utf-8") as f:
            search_results_geojson = json.load(f)
        cls.products = SearchResult.from_geojson(search_results_geojson)
        cls.arguments = {
            "query": {"eo:cloud_cover": {"lte": "10"}},
            "filter": "latestIntersect",
        }
        cls.criteria = {
            "productType": "S2_MSI_L1C",
            "page": 1,
            "items_per_page": 1,
            "raise_errors": True,
            "cloudCover": "10",
        }
        cls.empty_products = SearchResult([])
        cls.empty_arguments = {}
        cls.empty_criteria = {}

        # mock os.environ to empty env
        cls.mock_os_environ = mock.patch.dict(os.environ, {}, clear=True)
        cls.mock_os_environ.start()

        # disable product types fetch
        os.environ["EODAG_EXT_PRODUCT_TYPES_CFG_FILE"] = ""

        # create a dictionary obj of a S2_MSI_L1C peps response search
        peps_resp_search_file = os.path.join(
            TEST_RESOURCES_PATH, "provider_responses", "peps_search.json"
        )
        with open(peps_resp_search_file, encoding="utf-8") as f:
            cls.peps_resp_search_json = json.load(f)

        # create a dictionary obj of a S2_MSI_L2A earth_search response search
        earth_search_resp_search_file = os.path.join(
            TEST_RESOURCES_PATH, "provider_responses", "earth_search_search.json"
        )
        with open(earth_search_resp_search_file, encoding="utf-8") as f:
            cls.earth_search_resp_search_json = json.load(f)

    @classmethod
    def tearDownClass(cls):
        super(TestStacCore, cls).tearDownClass()
        # stop Mock and remove tmp config dir
        cls.expanduser_mock.stop()
        cls.tmp_home_dir.cleanup()
        # stop os.environ
        cls.mock_os_environ.stop()

    def test_crunch_products_unknown_cruncher_raise_error(self):
        """crunch_products must raise a ValidationError if an unknown cruncher is given"""
        with self.assertRaises(ValidationError) as context:
            self.rest_core.crunch_products(self.products, "unknown_cruncher")
        self.assertTrue("Unknown crunch name" in str(context.exception))

    def test_crunch_products_missing_additional_parameters_raise_error(self):
        """crunch_products must raise a ValidationError if additional parameters are required by the cruncher"""
        with self.assertRaises(ValidationError) as context:
            self.rest_core.crunch_products(self.products, "filterLatestByName")
        self.assertTrue("require additional parameters" in str(context.exception))

    def test_crunch_products_filter_misuse_raise_error(self):
        """crunch_products must raise a ValidationError if the cruncher is not used correctly"""
        with self.assertRaises(ValidationError):
            self.rest_core.crunch_products(
                self.products,
                "filterLatestByName",
                **{"name_pattern": "MisconfiguredError"},
            )

    def test_crunch_products(self):
        """crunch_products returns a SearchResult corresponding to the filter"""
        products_filtered = self.rest_core.crunch_products(
            self.products,
            "filterLatestByName",
            **{"name_pattern": r"S2[AB]_MSIL1C_20(?P<tileid>\d{6}).*T21NY.*"},
        )
        self.assertNotEqual(self.products, products_filtered)

    @mock.patch(
        "eodag.rest.core.eodag_api.list_product_types",
        autospec=True,
        return_value=[{"ID": "S2_MSI_L1C", "abstract": "test"}],
    )
    def test_format_product_types(self, list_pt: Mock):
        """format_product_types must return a string representation of the product types"""
        product_types = self.rest_core.eodag_api.list_product_types(
            fetch_providers=False
        )
        with pytest.warns(
            DeprecationWarning,
            match="Call to deprecated function/method format_product_types",
        ):
            self.assertEqual(
                self.rest_core.format_product_types(product_types),
                "* *__S2_MSI_L1C__*: test",
            )

    @mock.patch(
        "eodag.rest.core.eodag_api.list_product_types",
        autospec=True,
        return_value=[{"ID": "S2_MSI_L1C"}],
    )
    def test_detailled_collections_list(self, list_pt):
        """get_detailled_collections_list returned list is non-empty"""
        self.assertTrue(self.rest_core.get_detailled_collections_list())
        self.assertTrue(list_pt.called)

    def test_get_geometry(self):
        pass  # TODO

    def test_home_page_content(self):
        """get_home_page_content runs without any error"""
        with pytest.warns(
            DeprecationWarning,
            match="Call to deprecated function/method get_home_page_content",
        ):
            self.rest_core.get_home_page_content("http://127.0.0.1/")

    async def test_get_stac_catalogs(self):
        """get_stac_catalogs runs without any error"""

        await self.rest_core.get_stac_catalogs(mock_request("/"), url="")

    async def test_get_stac_collection(self):
        """get_collection runs without any error"""
        r = await self.rest_core.get_collection(
            mock_request("/"), collection_id="S2_MSI_L1C"
        )
        self.assertIsNotNone(r)
        self.assertEqual(8, len(r["providers"]))
        self.assertEqual(1, r["providers"][0]["priority"])
        self.assertEqual("peps", r["providers"][0]["name"])
        self.assertEqual(["host"], r["providers"][0]["roles"])
        self.assertEqual("https://peps.cnes.fr", r["providers"][0]["url"])
        self.assertTrue(
            r["providers"][0]["description"].startswith(
                'The PEPS platform, the French "mirror site"'
            )
        )

    async def test_get_stac_collections(self):
        """get_stac_collections runs without any error"""
        await self.rest_core.all_collections(mock_request("/"))

    async def test_all_collections_with_various_params(self):
        test_cases = [
            ({"bbox": "5,5,15,15"}, ["test_id"], "bbox intersects"),
            ({"bbox": "-10,-10,-5,-5"}, [], "bbox does not intersect"),
            ({"q": "test"}, ["test_id"], "free text query matches"),
            ({"q": "nomatch"}, [], "free text query does not match"),
            ({"platform": "test_platform"}, ["test_id"], "platform matches"),
            ({"platform": "other_platform"}, [], "platform does not match"),
            ({"instrument": "test_instrument"}, ["test_id"], "instrument matches"),
            ({"instrument": "other_instrument"}, [], "instrument does not match"),
            (
                {"constellation": "test_constellation"},
                ["test_id"],
                "constellation matches",
            ),
            (
                {"constellation": "other_constellation"},
                [],
                "constellation does not match",
            ),
            (
                {"datetime": "2020-01-01T00:00:00Z/2020-12-31T23:59:59Z"},
                ["test_id"],
                "datetime matches",
            ),
            (
                {"datetime": "1990-01-01T00:00:00Z/1990-12-31T23:59:59Z"},
                [],
                "datetime does not match",
            ),
            (
                {"bbox": "5,5,15,15", "q": "test", "platform": "test_platform"},
                ["test_id"],
                "multiple params match",
            ),
            (
                {"bbox": "5,5,15,15", "q": "nomatch", "platform": "test_platform"},
                [],
                "multiple params, one does not match",
            ),
        ]

        fake_product_type = {
            "ID": "test_id",
            "title": "Test Collection",
            "description": "A test collection",
            "keywords": ["test"],
            "license": "test-license",
            "platform": "test_platform",
            "instrument": "test_instrument",
            "constellation": "test_constellation",
        }

        # This is the external STAC collection, which provides the bbox for filtering
        fake_ext_stac_collection = {
            "extent": {"spatial": {"bbox": [[0.0, 0.0, 10.0, 10.0]]}},
        }

        def fake_guess_product_type(
            free_text=None,
            platformSerialIdentifier=None,
            instrument=None,
            platform=None,
            productType=None,
            missionStartDate=None,
            missionEndDate=None,
        ):
            if free_text is not None:
                return ["test_id"] if free_text == "test" else []
            if platformSerialIdentifier is not None:
                return (
                    ["test_id"] if platformSerialIdentifier == "test_platform" else []
                )
            if instrument is not None:
                return ["test_id"] if instrument == "test_instrument" else []
            if platform is not None:
                return ["test_id"] if platform == "test_constellation" else []
            if productType is not None:
                return ["test_id"] if productType == "test_id" else []
            if missionStartDate or missionEndDate:
                if missionStartDate and missionStartDate.startswith("2020"):
                    return ["test_id"]
                return []
            return ["test_id"]

        import inspect

        from eodag.rest.stac import StacCollection

        async def cached_side_effect(func, key, req):
            if inspect.iscoroutinefunction(func):
                return await func()
            return func()

        with (
            mock.patch(
                "eodag.rest.core.eodag_api.list_product_types",
                return_value=[fake_product_type],
            ),
            mock.patch(
                "eodag.rest.core.format_dict_items", side_effect=lambda x, **kwargs: x
            ),
            mock.patch("eodag.rest.core.cached", side_effect=cached_side_effect),
            mock.patch(
                "eodag.rest.core.eodag_api.guess_product_type",
                side_effect=fake_guess_product_type,
            ),
            mock.patch(
                "eodag.rest.core.SearchPostRequest.validate_bbox",
                side_effect=lambda bbox: None,
            ),
            mock.patch.object(
                StacCollection,
                "ext_stac_collections",
                {"test_id": fake_ext_stac_collection},
            ),
        ):
            for params, expected_ids, desc in test_cases:
                request = mock_request("http://testserver/collections")
                params_copy = params.copy()
                result = await self.rest_core.all_collections(
                    request=request, **params_copy
                )

                self.assertIn("collections", result, f"Failed: {desc}")
                ids = [col["id"] for col in result["collections"]]
                self.assertEqual(ids, expected_ids, f"Failed: {desc} (params={params})")

    def test_get_stac_conformance(self):
        """get_stac_conformance runs without any error"""
        self.rest_core.get_stac_conformance()

    def test_get_stac_extension_oseo(self):
        """get_stac_extension_oseo runs without any error"""
        self.rest_core.get_stac_extension_oseo(url="")

    def test_search_bbox(self):
        pass  # TODO

    def test_search_product_by_id(self):
        pass  # TODO

    @mock.patch(
        "eodag.plugins.search.qssearch.PostJsonSearch._request",
        autospec=True,
    )
    async def test_search_stac_items_with_stac_providers(self, mock__request: Mock):
        """search_stac_items runs without any error with stac providers"""
        # mock the PostJsonSearch request with the S2_MSI_L1C earth_search response search dictionary
        mock__request.return_value = mock.Mock()
        mock__request.return_value.json.return_value = (
            self.earth_search_resp_search_json
        )
        self.rest_core.eodag_api.set_preferred_provider("peps")

        response = await self.rest_core.search_stac_items(
            request=mock_request("http://foo/search"),
            search_request=SearchPostRequest.model_validate(
                {"collections": "S2_MSI_L1C", "provider": "earth_search"}
            ),
        )

        mock__request.assert_called()

        # check that default assets have been added to the response
        self.assertTrue(
            "downloadLink", "thumbnail" in response["features"][0]["assets"].keys()
        )
        # check that assets from the provider response search are reformatted in the response
        product_id = self.earth_search_resp_search_json["features"][0]["id"]
        for k in self.earth_search_resp_search_json["features"][0]["assets"]:
            self.assertIn(k, response["features"][0]["assets"].keys())
            if k == "thumbnail":
                self.assertTrue(response["features"][0]["assets"][k]["href"])
                continue
            # check asset server-mode download link
            self.assertEqual(
                response["features"][0]["assets"][k]["href"],
                f"http://foo/collections/S2_MSI_L1C/items/{product_id}/download/{k}?provider=earth_search",
            )
            # check asset origin download link
            self.assertEqual(
                response["features"][0]["assets"][k]["alternate"]["origin"]["href"],
                self.earth_search_resp_search_json["features"][0]["assets"][k]["href"],
            )
        # preferred provider should not be changed
        self.assertEqual("peps", self.rest_core.eodag_api.get_preferred_provider()[0])

    @mock.patch(
        "eodag.plugins.search.qssearch.QueryStringSearch._request",
        autospec=True,
    )
    async def test_search_stac_items_with_non_stac_providers(self, mock__request: Mock):
        """search_stac_items runs without any error with non-stac providers"""
        # mock the QueryStringSearch request with the S2_MSI_L1C peps response search dictionary
        mock__request.return_value = mock.Mock()
        mock__request.return_value.json.return_value = self.peps_resp_search_json

        response = await self.rest_core.search_stac_items(
            request=mock_request("http://foo/search"),
            search_request=SearchPostRequest.model_validate({"provider": "peps"}),
            catalogs=["S2_MSI_L1C"],
        )

        mock__request.assert_called()

        # check that default assets have been added to the response
        self.assertTrue(
            "downloadLink", "thumbnail" in response["features"][0]["assets"].keys()
        )
        # check that no other asset have also been added to the response
        self.assertEqual(len(response["features"][0]["assets"]), 2)

    @mock.patch(
        "eodag.plugins.search.qssearch.QueryStringSearch._request",
        autospec=True,
    )
    async def test_search_stac_items_get(self, mock__request: Mock):
        """search_stac_items runs with GET method"""
        # mock the QueryStringSearch request with the S2_MSI_L1C peps response search dictionary
        mock__request.return_value = mock.Mock()
        mock__request.return_value.json.return_value = self.peps_resp_search_json

        response = await self.rest_core.search_stac_items(
            request=mock_request("http://foo/search?collections=S2_MSI_L1C"),
            search_request=SearchPostRequest.model_validate(
                {"collections": ["S2_MSI_L1C"]}
            ),
        )

        mock__request.assert_called()

        next_link = [link for link in response["links"] if link["rel"] == "next"][0]

        self.assertEqual(
            next_link,
            {
                "method": "GET",
                "rel": "next",
                "href": "http://foo/search?collections=S2_MSI_L1C&page=2",
                "title": "Next page",
                "type": "application/geo+json",
            },
        )

    @mock.patch(
        "eodag.plugins.search.qssearch.QueryStringSearch._request",
        autospec=True,
    )
    async def test_search_stac_items_post(self, mock__request: Mock):
        """search_stac_items runs with GET method"""
        # mock the QueryStringSearch request with the S2_MSI_L1C peps response search dictionary
        mock__request.return_value = mock.Mock()
        mock__request.return_value.json.return_value = self.peps_resp_search_json

        response = await self.rest_core.search_stac_items(
            request=mock_request(
                url="http://foo/search",
                method="POST",
                body={"collections": ["S2_MSI_L1C"], "page": "2"},
            ),
            search_request=SearchPostRequest.model_validate(
                {"collections": ["S2_MSI_L1C"], "page": "2"}
            ),
        )

        mock__request.assert_called()

        next_link = [link for link in response["links"] if link["rel"] == "next"][0]

        self.assertEqual(
            next_link,
            {
                "method": "POST",
                "rel": "next",
                "href": "http://foo/search",
                "title": "Next page",
                "type": "application/geo+json",
                "body": {"collections": ["S2_MSI_L1C"], "page": 3},
            },
        )

    @mock.patch(
        "eodag.plugins.search.qssearch.QueryStringSearch._request",
        autospec=True,
    )
    def test_search_stac_items_special_characters(self, mock__request: Mock):
        """special characters in the id are quoted in the links"""
        # mock the QueryStringSearch request with the S2_MSI_L1C peps response search dictionary
        mock__request.return_value = mock.Mock()
        res = self.peps_resp_search_json
        res["features"][0]["properties"]["productIdentifier"] = "id,with,commas"
        res["features"][1]["properties"]["productIdentifier"] = "star*in*id"
        mock__request.return_value.json.return_value = res

        response = self.rest_core.search_stac_items(
            request=mock_request("http://foo/search"),
            search_request=SearchPostRequest.model_validate(
                {"provider": "peps", "collections": "S2_MSI_L1C"}
            ),
        )

        mock__request.assert_called()

        # check that default assets have been added to the response
        self.assertTrue(
            "downloadLink", "thumbnail" in response["features"][0]["assets"].keys()
        )
        # check that invalid characters have been quoted
        self.assertIn(",", response["features"][0]["id"])
        self.assertNotIn(",", response["features"][0]["assets"]["downloadLink"]["href"])
        self.assertNotIn(",", response["features"][0]["links"][0]["href"])
        self.assertIn("*", response["features"][1]["id"])
        self.assertNotIn("*", response["features"][1]["assets"]["downloadLink"]["href"])
        self.assertNotIn("*", response["features"][1]["links"][0]["href"])
        # check that no other asset have also been added to the response
        self.assertEqual(len(response["features"][0]["assets"]), 2)

    def test_get_templates_path(self):
        """get_templates_path returns an existing dir path"""
        with pytest.warns(
            DeprecationWarning,
            match="Call to deprecated function/method get_templates_path",
        ):
            self.assertTrue(os.path.isdir(self.rest_core.get_templates_path()))



# -*- coding: utf-8 -*-
# Copyright 2023, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SerializerFunctionWrapHandler,
    model_serializer,
)

from eodag.rest.types.eodag_search import EODAGSearch


class CollectionsSearchRequest(BaseModel):
    """Search args for GET collections"""

    model_config = ConfigDict(frozen=True)

    q: Optional[str] = Field(default=None, serialization_alias="free_text")
    platform: Optional[str] = Field(default=None)
    instrument: Optional[str] = Field(default=None)
    constellation: Optional[str] = Field(default=None)

    @model_serializer(mode="wrap")
    def _serialize(self, handler: SerializerFunctionWrapHandler) -> dict[str, Any]:
        dumped: dict[str, Any] = handler(self)
        return {EODAGSearch.to_eodag(k): v for k, v in dumped.items()}



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from functools import lru_cache
from typing import Annotated, Union

from pydantic import Field
from pydantic.functional_validators import BeforeValidator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Doc

from eodag.rest.constants import DEFAULT_MAXSIZE, DEFAULT_TTL


def str2liststr(raw: Union[str, list[str]]) -> list[str]:
    """Convert str to list[str]"""
    if isinstance(raw, list):
        return raw
    return raw.split(",")


class Settings(BaseSettings):
    """EODAG Server config"""

    # local cache config
    cache_ttl: int = Field(default=DEFAULT_TTL)
    cache_maxsize: int = Field(default=DEFAULT_MAXSIZE)

    debug: bool = False

    stac_api_title: str = "eodag-stac-api"
    stac_api_description: str = "STAC API provided by EODAG"
    stac_api_landing_id: str = "eodag-stac-api"

    origin_url_blacklist: Annotated[
        Union[str, list[str]],
        BeforeValidator(str2liststr),
        Doc(
            "Hide from clients items assets' alternative URLs starting with URLs from the list"
        ),
    ] = Field(default=[])

    model_config = SettingsConfigDict(
        env_prefix="EODAG_", extra="ignore", env_nested_delimiter="__"
    )

    @classmethod
    @lru_cache(maxsize=1)
    def from_environment(cls) -> Settings:
        """Get settings"""
        return Settings()



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import importlib
import json
import os
import socket
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional, Union
from unittest.mock import Mock, call
from urllib.parse import quote_plus

import geojson
import httpx
import responses
from fastapi.testclient import TestClient
from shapely.geometry import box

from eodag.config import PluginConfig
from eodag.plugins.authentication.base import Authentication
from eodag.plugins.download.base import Download
from eodag.rest.config import Settings
from eodag.rest.types.queryables import StacQueryables
from eodag.utils import USER_AGENT, MockResponse
from eodag.utils.exceptions import RequestError, TimeOutError, ValidationError
from tests import mock, temporary_environment
from tests.context import (
    DEFAULT_ITEMS_PER_PAGE,
    DEFAULT_SEARCH_TIMEOUT,
    OFFLINE_STATUS,
    ONLINE_STATUS,
    TEST_RESOURCES_PATH,
    AuthenticationError,
    NotAvailableError,
    PluginImplementationError,
    SearchResult,
)


# AF_UNIX socket not supported on windows yet, see https://github.com/python/cpython/issues/77589
@unittest.skipIf(
    not hasattr(socket, "AF_UNIX"), "AF_UNIX socket not supported on this OS (windows)"
)
class RequestTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(RequestTestCase, cls).setUpClass()

        cls.tested_product_type = "S2_MSI_L1C"

        # Mock home and eodag conf directory to tmp dir
        cls.tmp_home_dir = TemporaryDirectory()
        cls.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, return_value=cls.tmp_home_dir.name
        )
        cls.expanduser_mock.start()

        # mock os.environ to empty env
        cls.mock_os_environ = mock.patch.dict(os.environ, {}, clear=True)
        cls.mock_os_environ.start()

        # disable product types fetch
        os.environ["EODAG_EXT_PRODUCT_TYPES_CFG_FILE"] = ""

        # load fake credentials to prevent providers needing auth for search to be pruned
        os.environ["EODAG_CFG_FILE"] = os.path.join(
            TEST_RESOURCES_PATH, "wrong_credentials_conf.yml"
        )

        # import after having mocked home_dir because it launches http server (and EODataAccessGateway)
        # reload eodag.rest.core to prevent eodag_api cache conflicts
        import eodag.rest.core

        importlib.reload(eodag.rest.core)
        from eodag.rest import server as eodag_http_server

        cls.eodag_http_server = eodag_http_server

    @classmethod
    def tearDownClass(cls):
        super(RequestTestCase, cls).tearDownClass()
        # stop os.environ
        cls.mock_os_environ.stop()

        # stop Mock and remove tmp config dir
        cls.expanduser_mock.stop()
        cls.tmp_home_dir.cleanup()

    def setUp(self):
        self.app = TestClient(self.eodag_http_server.app)

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    def test_route(self, mock_auth_session_request):
        result = self._request_valid("/", check_links=False)

        # check links (root specfic)
        self.assertIsInstance(result, dict)
        self.assertIn("links", result, f"links not found in {str(result)}")
        self.assertIsInstance(result["links"], list)
        links = result["links"]

        known_rel = [
            "self",
            "root",
            "parent",
            "child",
            "items",
            "service-desc",
            "service-doc",
            "conformance",
            "search",
            "data",
        ]
        required_links_rel = ["self"]

        for link in links:
            # known relations
            self.assertIn(link["rel"], known_rel)
            # must start with app base-url
            assert link["href"].startswith(str(self.app.base_url))
            if link["rel"] != "search":
                # must be valid
                self._request_valid_raw(link["href"])
            else:
                # missing collection
                self._request_not_valid(link["href"])

            if link["rel"] in required_links_rel:
                required_links_rel.remove(link["rel"])

        # required relations
        self.assertEqual(
            len(required_links_rel),
            0,
            f"missing {required_links_rel} relation(s) in {links}",
        )

    def test_forward(self):
        response = self.app.get("/", follow_redirects=True)
        self.assertEqual(200, response.status_code)
        resp_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(resp_json["links"][0]["href"], "http://testserver/")

        response = self.app.get(
            "/", follow_redirects=True, headers={"Forwarded": "host=foo;proto=https"}
        )
        self.assertEqual(200, response.status_code)
        resp_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(resp_json["links"][0]["href"], "https://foo/")

        response = self.app.get(
            "/",
            follow_redirects=True,
            headers={"X-Forwarded-Host": "bar", "X-Forwarded-Proto": "httpz"},
        )
        self.assertEqual(200, response.status_code)
        resp_json = json.loads(response.content.decode("utf-8"))
        self.assertEqual(resp_json["links"][0]["href"], "httpz://bar/")

    def test_liveness_probe(self):
        response = self.app.get("/_mgmt/ping")
        self.assertEqual(200, response.status_code)
        self.assertTrue(response.json()["success"])

    def mock_search_result(self):
        """generate eodag_api.search mock results"""
        search_result = SearchResult.from_geojson(
            {
                "features": [
                    {
                        "properties": {
                            "snowCover": None,
                            "resolution": None,
                            "completionTimeFromAscendingNode": "2018-02-16T00:12:14.035Z",
                            "keyword": {},
                            "productType": "OCN",
                            "downloadLink": (
                                "https://peps.cnes.fr/resto/collections/S1/"
                                "578f1768-e66e-5b86-9363-b19f8931cc7b/download"
                            ),
                            "eodag_provider": "peps",
                            "eodag_product_type": "S1_SAR_OCN",
                            "platformSerialIdentifier": "S1A",
                            "cloudCover": 0,
                            "title": "S1A_WV_OCN__2SSV_20180215T235323_"
                            "20180216T001213_020624_023501_0FD3",
                            "orbitNumber": 20624,
                            "instrument": "SAR-C SAR",
                            "abstract": None,
                            "eodag_search_intersection": {
                                "coordinates": [
                                    [
                                        [89.590721, 2.614019],
                                        [89.771805, 2.575546],
                                        [89.809341, 2.756323],
                                        [89.628258, 2.794767],
                                        [89.590721, 2.614019],
                                    ]
                                ],
                                "type": "Polygon",
                            },
                            "organisationName": None,
                            "startTimeFromAscendingNode": "2018-02-15T23:53:22.871Z",
                            "platform": None,
                            "sensorType": None,
                            "processingLevel": None,
                            "orbitType": None,
                            "topicCategory": None,
                            "orbitDirection": None,
                            "parentIdentifier": None,
                            "sensorMode": None,
                            "quicklook": None,
                            "storageStatus": ONLINE_STATUS,
                            "providerProperty": "foo",
                        },
                        "id": "578f1768-e66e-5b86-9363-b19f8931cc7b",
                        "type": "Feature",
                        "geometry": {
                            "coordinates": [
                                [
                                    [89.590721, 2.614019],
                                    [89.771805, 2.575546],
                                    [89.809341, 2.756323],
                                    [89.628258, 2.794767],
                                    [89.590721, 2.614019],
                                ]
                            ],
                            "type": "Polygon",
                        },
                    },
                    {
                        "properties": {
                            "snowCover": None,
                            "resolution": None,
                            "completionTimeFromAscendingNode": "2018-02-17T00:12:14.035Z",
                            "keyword": {},
                            "productType": "OCN",
                            "downloadLink": (
                                "https://peps.cnes.fr/resto/collections/S1/"
                                "578f1768-e66e-5b86-9363-b19f8931cc7c/download"
                            ),
                            "eodag_provider": "peps",
                            "eodag_product_type": "S1_SAR_OCN",
                            "platformSerialIdentifier": "S1A",
                            "cloudCover": 0,
                            "title": "S1A_WV_OCN__2SSV_20180216T235323_"
                            "20180217T001213_020624_023501_0FD3",
                            "orbitNumber": 20624,
                            "instrument": "SAR-C SAR",
                            "abstract": None,
                            "eodag_search_intersection": {
                                "coordinates": [
                                    [
                                        [89.590721, 2.614019],
                                        [89.771805, 2.575546],
                                        [89.809341, 2.756323],
                                        [89.628258, 2.794767],
                                        [89.590721, 2.614019],
                                    ]
                                ],
                                "type": "Polygon",
                            },
                            "organisationName": None,
                            "startTimeFromAscendingNode": "2018-02-16T23:53:22.871Z",
                            "platform": None,
                            "sensorType": None,
                            "processingLevel": None,
                            "orbitType": None,
                            "topicCategory": None,
                            "orbitDirection": None,
                            "parentIdentifier": None,
                            "sensorMode": None,
                            "quicklook": None,
                            "storageStatus": OFFLINE_STATUS,
                        },
                        "id": "578f1768-e66e-5b86-9363-b19f8931cc7c",
                        "type": "Feature",
                        "geometry": {
                            "coordinates": [
                                [
                                    [89.590721, 2.614019],
                                    [89.771805, 2.575546],
                                    [89.809341, 2.756323],
                                    [89.628258, 2.794767],
                                    [89.590721, 2.614019],
                                ]
                            ],
                            "type": "Polygon",
                        },
                    },
                ],
                "type": "FeatureCollection",
            }
        )
        config = PluginConfig()
        config.priority = 0
        for p in search_result:
            p.downloader = Download("peps", config)
            p.downloader_auth = Authentication("peps", config)
        search_result.number_matched = len(search_result)
        return search_result

    @mock.patch("eodag.rest.core.eodag_api.search", autospec=True)
    def _request_valid_raw(
        self,
        url: str,
        mock_search: Mock,
        expected_search_kwargs: Union[
            list[dict[str, Any]], dict[str, Any], None
        ] = None,
        method: str = "GET",
        post_data: Optional[Any] = None,
        search_call_count: Optional[int] = None,
        search_result: Optional[SearchResult] = None,
        expected_status_code: int = 200,
    ) -> httpx.Response:
        if search_result:
            mock_search.return_value = search_result
        else:
            mock_search.return_value = self.mock_search_result()
        response = self.app.request(
            method,
            url,
            json=post_data,
            follow_redirects=True,
            headers={"Content-Type": "application/json"} if method == "POST" else {},
        )

        if search_call_count is not None:
            self.assertEqual(mock_search.call_count, search_call_count)

        if (
            expected_search_kwargs is not None
            and search_call_count is not None
            and search_call_count > 1
        ):
            self.assertIsInstance(
                expected_search_kwargs,
                list,
                "expected_search_kwargs must be a list if search_call_count > 1",
            )
            for single_search_kwargs in expected_search_kwargs:
                mock_search.assert_any_call(**single_search_kwargs)
        elif expected_search_kwargs is not None:
            mock_search.assert_called_once_with(**expected_search_kwargs)

        self.assertEqual(expected_status_code, response.status_code, response.text)

        return response

    def _request_valid(
        self,
        url: str,
        expected_search_kwargs: Union[
            list[dict[str, Any]], dict[str, Any], None
        ] = None,
        method: str = "GET",
        post_data: Optional[Any] = None,
        search_call_count: Optional[int] = None,
        check_links: bool = True,
        search_result: Optional[SearchResult] = None,
    ) -> Any:
        response = self._request_valid_raw(
            url,
            expected_search_kwargs=expected_search_kwargs,
            method=method,
            post_data=post_data,
            search_call_count=search_call_count,
            search_result=search_result,
        )

        # Assert response format is GeoJSON
        result = geojson.loads(response.content.decode("utf-8"))

        if check_links:
            self.assert_links_valid(result)

        return result

    def assert_links_valid(self, element: Any):
        """Checks that element links are valid"""
        self.assertIsInstance(element, dict)
        self.assertIn("links", element, f"links not found in {str(element)}")
        self.assertIsInstance(element["links"], list)
        links = element["links"]

        known_rel = [
            "self",
            "root",
            "parent",
            "child",
            "items",
            "service-desc",
            "service-doc",
            "conformance",
            "search",
            "data",
            "collection",
        ]
        required_links_rel = ["self", "root"]

        for link in links:
            # known relations
            self.assertIn(link["rel"], known_rel)
            # must start with app base-url
            assert link["href"].startswith(str(self.app.base_url))
            # HEAD must be valid
            self._request_valid_raw(link["href"], method="HEAD")
            # GET must be valid
            self._request_valid_raw(link["href"])

            if link["rel"] in required_links_rel:
                required_links_rel.remove(link["rel"])

        # required relations
        self.assertEqual(
            len(required_links_rel),
            0,
            f"missing {required_links_rel} relation(s) in {links}",
        )

    def _request_not_valid(
        self, url: str, method: str = "GET", post_data: Optional[Any] = None
    ) -> None:
        response = self.app.request(
            method,
            url,
            json=post_data,
            follow_redirects=True,
            headers={"Content-Type": "application/json"} if method == "POST" else {},
        )
        response_content = json.loads(response.content.decode("utf-8"))

        self.assertEqual(400, response.status_code)
        self.assertIn("description", response_content)

    def _request_not_found(self, url: str):
        response = self.app.get(url, follow_redirects=True)
        response_content = json.loads(response.content.decode("utf-8"))

        self.assertEqual(404, response.status_code)
        self.assertIn("description", response_content)
        self.assertIn("NotAvailableError", response_content["description"])

    def _request_accepted(self, url: str):
        response = self.app.get(url, follow_redirects=True)
        response_content = json.loads(response.content.decode("utf-8"))
        self.assertEqual(202, response.status_code)
        self.assertIn("description", response_content)
        self.assertIn("location", response_content)
        return response_content

    def test_request_params(self):
        self._request_not_valid(f"search?collections={self.tested_product_type}&bbox=1")
        self._request_not_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1"
        )
        self._request_not_valid(
            f"search?collections={self.tested_product_type}&bbox=0,,1"
        )
        self._request_not_valid(
            f"search?collections={self.tested_product_type}&bbox=a,43,1,44"
        )

        self._request_valid(
            f"search?collections={self.tested_product_type}",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1,44",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )

    def test_items_response(self):
        """Returned items properties must be mapped as expected"""
        resp_json = self._request_valid(
            f"search?collections={self.tested_product_type}",
        )
        res = resp_json.features
        self.assertEqual(len(res), 2)
        first_props = res[0]["properties"]
        self.assertCountEqual(
            res[0].keys(),
            [
                "type",
                "stac_version",
                "stac_extensions",
                "bbox",
                "collection",
                "links",
                "assets",
                "id",
                "geometry",
                "properties",
            ],
        )
        self.assertEqual(len(first_props["providers"]), 1)
        self.assertCountEqual(
            first_props["providers"][0].keys(),
            ["name", "description", "roles", "url", "priority"],
        )
        self.assertEqual(first_props["providers"][0]["name"], "peps")
        self.assertEqual(first_props["providers"][0]["roles"], ["host"])
        self.assertEqual(first_props["providers"][0]["url"], "https://peps.cnes.fr")
        self.assertEqual(first_props["datetime"], "2018-02-15T23:53:22.871Z")
        self.assertEqual(first_props["start_datetime"], "2018-02-15T23:53:22.871Z")
        self.assertEqual(first_props["end_datetime"], "2018-02-16T00:12:14.035Z")
        self.assertEqual(first_props["license"], "other")
        self.assertEqual(first_props["platform"], "S1A")
        self.assertEqual(first_props["instruments"], ["SAR-C SAR"])
        self.assertEqual(first_props["eo:cloud_cover"], 0)
        self.assertEqual(first_props["sat:absolute_orbit"], 20624)
        self.assertEqual(first_props["sar:product_type"], "OCN")
        self.assertEqual(first_props["order:status"], "succeeded")
        self.assertEqual(res[0]["assets"]["downloadLink"]["storage:tier"], "ONLINE")
        self.assertEqual(res[1]["assets"]["downloadLink"]["storage:tier"], "OFFLINE")
        self.assertEqual(res[1]["properties"]["order:status"], "orderable")

    def test_not_found(self):
        """A request to eodag server with a not supported product type must return a 404 HTTP error code"""
        self._request_not_found("search?collections=ZZZ&bbox=0,43,1,44")

    @mock.patch(
        "eodag.rest.core.eodag_api.search",
        autospec=True,
        side_effect=AuthenticationError("you are not authorized"),
    )
    def test_auth_error(self, mock_search: Mock):
        """A request to eodag server raising a Authentication error must return a 500 HTTP error code"""

        with self.assertLogs(level="ERROR") as cm_logs:
            response = self.app.get(
                f"search?collections={self.tested_product_type}", follow_redirects=True
            )
            response_content = json.loads(response.content.decode("utf-8"))

            self.assertIn("description", response_content)
            self.assertIn("Internal server error", response_content["description"])
            self.assertIn("AuthenticationError", str(cm_logs.output))
            self.assertIn("you are not authorized", str(cm_logs.output))

        self.assertEqual(500, response.status_code)

    @mock.patch(
        "eodag.rest.core.eodag_api.search",
        autospec=True,
        side_effect=TimeOutError("too long"),
    )
    def test_timeout_error(self, mock_search: Mock):
        """A request to eodag server raising a timeout error must return a 504 HTTP error code"""
        with self.assertLogs(level="ERROR") as cm_logs:
            response = self.app.get(
                f"search?collections={self.tested_product_type}", follow_redirects=True
            )
            response_content = json.loads(response.content.decode("utf-8"))

            self.assertIn("description", response_content)
            self.assertIn("TimeOutError", response_content["description"])
            self.assertIn("too long", response_content["description"])
            self.assertIn("TimeOutError", str(cm_logs.output))
            self.assertIn("too long", str(cm_logs.output))

        self.assertEqual(504, response.status_code)

    @mock.patch(
        "eodag.rest.core.eodag_api.search",
        autospec=True,
        side_effect=PluginImplementationError("some error"),
    )
    def test_plugin_implementation_error(self, mock_search: Mock):
        """A request to eodag server raising a plugin implementation error must return a 500 HTTP error code"""
        with self.assertLogs(level="ERROR") as cm_logs:
            response = self.app.get(
                f"search?collections={self.tested_product_type}", follow_redirects=True
            )
            response_content = json.loads(response.content.decode("utf-8"))

            self.assertIn("description", response_content)
            self.assertIn("Internal server error", response_content["description"])
            self.assertIn("PluginImplementationError", str(cm_logs.output))
            self.assertIn("some error", str(cm_logs.output))

        self.assertEqual(500, response.status_code)

    def test_filter(self):
        """latestIntersect filter should only keep the latest products once search area is fully covered"""
        result1 = self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=89.65,2.65,89.7,2.7",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                geom=box(89.65, 2.65, 89.7, 2.7, ccw=False),
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                raise_errors=False,
                count=True,
            ),
        )
        self.assertEqual(len(result1.features), 2)
        result2 = self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=89.65,2.65,89.7,2.7&crunch=filterLatestIntersect",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                geom=box(89.65, 2.65, 89.7, 2.7, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )
        # only one product is returned with filter=latestIntersect
        self.assertEqual(len(result2.features), 1)

    def test_date_search(self):
        """Search through eodag server /search endpoint using dates filering should return a valid response"""
        self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1,44&datetime=2018-01-20/2018-01-25",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                end="2018-01-25T00:00:00.000Z",
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1,44&datetime=2018-01-20/..",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1,44&datetime=../2018-01-25",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                end="2018-01-25T00:00:00.000Z",
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            f"search?collections={self.tested_product_type}&bbox=0,43,1,44&datetime=2018-01-20",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                end="2018-01-20T00:00:00.000Z",
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )

    def test_date_search_from_items(self):
        """Search through eodag server collection/items endpoint using dates filering should return a valid response"""
        self._request_valid(
            f"collections/{self.tested_product_type}/items?bbox=0,43,1,44",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            f"collections/{self.tested_product_type}/items?bbox=0,43,1,44&datetime=2018-01-20/2018-01-25",
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                end="2018-01-25T00:00:00.000Z",
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )

    def test_search_item_id_from_collection(self):
        """Search by id through eodag server /collection endpoint should return a valid response"""
        self._request_valid(
            f"collections/{self.tested_product_type}/items/foo",
            expected_search_kwargs={
                "id": "foo",
                "productType": self.tested_product_type,
                "provider": None,
            },
        )

    def test_collection(self):
        """Requesting a collection through eodag server should return a valid response"""
        result = self._request_valid(f"collections/{self.tested_product_type}")
        self.assertEqual(result["id"], self.tested_product_type)
        for link in result["links"]:
            self.assertIn(link["rel"], ["self", "root", "items"])

    def test_cloud_cover_post_search(self):
        """POST search with cloudCover filtering through eodag server should return a valid response"""
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "bbox": [0, 43, 1, 44],
                "query": {"eo:cloud_cover": {"lte": 10}},
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                cloudCover=10,
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )

    def test_intersects_post_search(self):
        """POST search with intersects filtering through eodag server should return a valid response"""
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "intersects": {
                    "type": "Polygon",
                    "coordinates": [[[0, 43], [0, 44], [1, 44], [1, 43], [0, 43]]],
                },
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                geom=box(0, 43, 1, 44, ccw=False),
                raise_errors=False,
                count=True,
            ),
        )

    def test_date_post_search(self):
        """POST search with datetime filtering through eodag server should return a valid response"""
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "datetime": "2018-01-20/2018-01-25",
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                end="2018-01-25T00:00:00.000Z",
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "datetime": "2018-01-20/..",
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "datetime": "../2018-01-25",
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                end="2018-01-25T00:00:00.000Z",
                raise_errors=False,
                count=True,
            ),
        )
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "datetime": "2018-01-20",
            },
            expected_search_kwargs=dict(
                productType=self.tested_product_type,
                page=1,
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                start="2018-01-20T00:00:00.000Z",
                end="2018-01-20T00:00:00.000Z",
                raise_errors=False,
                count=True,
            ),
        )

    def test_ids_post_search(self):
        """POST search with ids filtering through eodag server should return a valid response"""
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "collections": [self.tested_product_type],
                "ids": ["foo", "bar"],
            },
            search_call_count=2,
            expected_search_kwargs=[
                {
                    "provider": None,
                    "id": "foo",
                    "productType": self.tested_product_type,
                },
                {
                    "provider": None,
                    "id": "bar",
                    "productType": self.tested_product_type,
                },
            ],
        )

    @mock.patch("eodag.rest.core.eodag_api.search", autospec=True)
    def test_provider_prefix_post_search(self, mock_search):
        """provider prefixes should be removed from query parameters"""
        post_data = {
            "collections": ["ERA5_SL"],
            "provider": "cop_cds",
            "query": {
                "cop_cds:month": {"eq": "10"},
                "cop_cds:year": {"eq": "2010"},
                "cop_cds:day": {"eq": "10"},
            },
        }
        mock_search.return_value = SearchResult.from_geojson(
            {
                "features": [],
                "type": "FeatureCollection",
            }
        )
        self.app.request(
            method="POST",
            url="search",
            json=post_data,
            follow_redirects=True,
            headers={"Content-Type": "application/json"},
        )
        expected_search_kwargs = dict(
            productType="ERA5_SL",
            page=1,
            items_per_page=DEFAULT_ITEMS_PER_PAGE,
            month="10",
            year="2010",
            day="10",
            raise_errors=False,
            count=True,
            provider="cop_cds",
        )
        mock_search.assert_called_once_with(**expected_search_kwargs)

    def test_search_response_contains_pagination_info(self):
        """Responses to valid search requests must return a geojson with pagination info in properties"""
        response = self._request_valid(f"search?collections={self.tested_product_type}")
        self.assertIn("numberMatched", response)
        self.assertIn("numberReturned", response)

    def test_search_provider_in_downloadlink(self):
        """Search through eodag server and check that specified provider appears in downloadLink"""
        # with provider (get)
        response = self._request_valid(
            f"search?collections={self.tested_product_type}&provider=peps"
        )
        response_items = [f for f in response["features"]]
        self.assertTrue(
            all(
                [
                    i["assets"]["downloadLink"]["href"].endswith(
                        "download?provider=peps"
                    )
                    for i in response_items
                ]
            )
        )
        # with provider (post)
        response = self._request_valid(
            "search",
            method="POST",
            post_data={"collections": [self.tested_product_type], "provider": "peps"},
        )
        response_items = [f for f in response["features"]]
        self.assertTrue(
            all(
                [
                    i["assets"]["downloadLink"]["href"].endswith(
                        "download?provider=peps"
                    )
                    for i in response_items
                ]
            )
        )
        # without provider
        response = self._request_valid(f"search?collections={self.tested_product_type}")
        response_items = [f for f in response["features"]]
        self.assertTrue(
            all(
                [
                    i["assets"]["downloadLink"]["href"].endswith(
                        "download?provider=peps"
                    )
                    for i in response_items
                ]
            )
        )

    def test_search_results_with_errors(self):
        """Search through eodag server must not display provider's error if it's not empty result"""
        errors = [
            ("usgs", Exception("foo error")),
            ("aws_eos", Exception("boo error")),
        ]
        search_results = self.mock_search_result()
        search_results.errors.extend(errors)

        self._request_valid(
            f"search?collections={self.tested_product_type}",
            search_result=search_results,
        )

    @mock.patch(
        "eodag.rest.core.eodag_api.search",
        autospec=True,
    )
    def test_search_no_results_with_errors(self, mock_search):
        """Search through eodag server must display provider's error if it's empty result"""
        req_err = RequestError("Request error message with status code")
        req_err.status_code = 418
        errors = [
            ("usgs", Exception("Generic exception", "Details of the error")),
            ("earth_search", TimeOutError("Timeout message")),
            ("peps", req_err),
            ("creodias", AuthenticationError("Authentication message")),
            (
                "cop_dataspace",
                ValidationError(
                    "Validation message: startTimeFromAscendingNode, modificationDate",
                    {"startTimeFromAscendingNode", "modificationDate"},
                ),
            ),
        ]
        expected_response = {
            "errors": [
                {
                    "provider": "usgs",
                    "error": "Exception",
                    "message": "Generic exception",
                    "detail": "Details of the error",
                    "status_code": 500,
                },
                {
                    "provider": "earth_search",
                    "error": "TimeOutError",
                    "message": "Timeout message",
                    "status_code": 504,
                },
                {
                    "provider": "peps",
                    "error": "RequestError",
                    "message": "Request error message with status code",
                    "status_code": 418,
                },
                {
                    "provider": "creodias",
                    "error": "AuthenticationError",
                    "message": "Internal server error: please contact the administrator",
                    "status_code": 500,
                },
                {
                    "provider": "cop_dataspace",
                    "error": "ValidationError",
                    "message": "Validation message: start_datetime, updated",
                    "detail": {"startTimeFromAscendingNode", "modificationDate"},
                    "status_code": 400,
                },
            ]
        }
        mock_search.return_value = SearchResult([], 0, errors)

        response = self.app.request(
            "GET",
            "search?collections=S2_MSI_L1C",
            json=None,
            follow_redirects=True,
            headers={},
        )
        response_content = json.loads(response.content.decode("utf-8"))

        self.assertEqual(400, response.status_code)
        for record in response_content["errors"]:
            if "detail" in record and "{" in record["detail"]:
                record["detail"] = (
                    record["detail"].replace("{", "").replace("}", "").replace("'", "")
                )
                record["detail"] = set(s.strip() for s in record["detail"].split(","))
        self.assertDictEqual(expected_response, response_content)

    def test_assets_alt_url_blacklist(self):
        """Search through eodag server must not have alternate link if in blacklist"""
        # no blacklist
        response = self._request_valid(f"search?collections={self.tested_product_type}")
        response_items = [f for f in response["features"]]
        self.assertTrue(
            all(["alternate" in i["assets"]["downloadLink"] for i in response_items]),
            "alternate links are missing",
        )

        # with blacklist
        try:
            Settings.from_environment.cache_clear()
            with temporary_environment(
                EODAG_ORIGIN_URL_BLACKLIST="https://peps.cnes.fr"
            ):
                response = self._request_valid(
                    f"search?collections={self.tested_product_type}"
                )
                response_items = [f for f in response["features"]]
                self.assertTrue(
                    all(
                        [
                            "alternate" not in i["assets"]["downloadLink"]
                            for i in response_items
                        ]
                    ),
                    "alternate links were not removed",
                )
        finally:
            Settings.from_environment.cache_clear()

    @mock.patch(
        "eodag.rest.core.eodag_api.guess_product_type", autospec=True, return_value=[]
    )
    @mock.patch(
        "eodag.rest.core.eodag_api.list_product_types",
        autospec=True,
        return_value=[
            {"_id": "S2_MSI_L1C", "ID": "S2_MSI_L1C", "title": "SENTINEL2 Level-1C"},
            {"_id": "S2_MSI_L2A", "ID": "S2_MSI_L2A"},
        ],
    )
    def test_list_product_types_ok(self, list_pt: Mock, guess_pt: Mock):
        """A simple request for product types with(out) a provider must succeed"""
        for url in ("/collections",):
            r = self.app.get(url)
            self.assertTrue(list_pt.called)
            self.assertEqual(200, r.status_code)
            self.assertListEqual(
                ["S2_MSI_L1C", "S2_MSI_L2A"],
                [
                    col["id"]
                    for col in json.loads(r.content.decode("utf-8")).get(
                        "collections", []
                    )
                ],
            )

        guess_pt.return_value = ["S2_MSI_L1C"]
        url = "/collections?instrument=MSI"
        r = self.app.get(url)
        self.assertTrue(guess_pt.called)
        self.assertTrue(list_pt.called)
        self.assertEqual(200, r.status_code)
        resp_json = json.loads(r.content.decode("utf-8"))
        self.assertListEqual(
            ["S2_MSI_L1C"],
            [col["id"] for col in resp_json.get("collections", [])],
        )
        self.assertEqual(resp_json["collections"][0]["title"], "SENTINEL2 Level-1C")

    @mock.patch(
        "eodag.rest.core.eodag_api.list_product_types",
        autospec=True,
        return_value=[
            {"_id": "S2_MSI_L1C", "ID": "S2_MSI_L1C"},
            {"_id": "S2_MSI_L2A", "ID": "S2_MSI_L2A"},
        ],
    )
    def test_list_product_types_nok(self, list_pt: Mock):
        """A request for product types with a not supported filter must return all product types"""
        url = "/collections?gibberish=gibberish"
        r = self.app.get(url)
        self.assertTrue(list_pt.called)
        self.assertEqual(200, r.status_code)
        self.assertListEqual(
            ["S2_MSI_L1C", "S2_MSI_L2A"],
            [
                col["id"]
                for col in json.loads(r.content.decode("utf-8")).get("collections", [])
            ],
        )

    @mock.patch(
        "eodag.plugins.authentication.base.Authentication.authenticate",
        autospec=True,
    )
    @mock.patch(
        "eodag.plugins.download.base.Download._stream_download_dict",
        autospec=True,
    )
    @mock.patch(
        "eodag.rest.core.eodag_api.download",
        autospec=True,
    )
    def test_download_item_from_collection_no_stream(
        self, mock_download: Mock, mock_stream_download: Mock, mock_auth: Mock
    ):
        """Download through eodag server catalog should return a valid response"""
        # download should be performed locally then deleted if streaming is not available
        tmp_dl_dir = TemporaryDirectory()
        expected_file = f"{tmp_dl_dir.name}.tar"
        Path(expected_file).touch()
        mock_download.return_value = expected_file
        mock_stream_download.side_effect = NotImplementedError()

        self._request_valid_raw(
            f"collections/{self.tested_product_type}/items/foo/download?provider=peps"
        )
        mock_download.assert_called_once()
        # downloaded file should have been immediatly deleted from the server
        assert not os.path.exists(
            expected_file
        ), f"File {expected_file} should have been deleted"

    @mock.patch(
        "eodag.plugins.authentication.base.Authentication.authenticate",
        autospec=True,
    )
    @mock.patch(
        "eodag.rest.core._order_and_update",
        autospec=True,
        side_effect=NotAvailableError(),
    )
    def test_download_item_orderable(
        self,
        mock_order_and_update: Mock,
        mock_auth: Mock,
    ):
        """Download orderable item through eodag server should order the item"""

        qs = '{"foo": "bar"}'
        self.app.request(
            "GET",
            f"collections/foo/items/FOO_ORDERABLE_13245/download?provider=cop_cds&_dc_qs={quote_plus(qs)}",
            json=None,
            follow_redirects=True,
            headers={},
        )

        mock_order_and_update.assert_called_once()
        self.assertEqual(
            mock_order_and_update.call_args[0][0].properties["qs"]["foo"], "bar"
        )

    @mock.patch(
        "eodag.plugins.authentication.base.Authentication.authenticate",
        autospec=True,
    )
    @mock.patch(
        "eodag.plugins.download.base.Download._stream_download_dict",
        autospec=True,
    )
    def test_error_handler_supports_request_exception_with_status_code_none(
        self, mock_stream_download: Mock, mock_auth: Mock
    ):
        """
        A RequestError with a status code set to None (the default value) should not
        crash the server. This test ensures that it doesn't crash the server and that a
        500 status code is returned to the user.
        """
        # Make _stream_download_dict reaise an exception object with a status_code
        # attribute set to None
        exception = RequestError()
        exception.status_code = None
        mock_stream_download.side_effect = exception

        response = self._request_valid_raw(
            f"collections/{self.tested_product_type}/items/foo/download?provider=peps",
            expected_status_code=500,
        )
        self.assertIn("RequestError", response.text)

    def test_root(self):
        """Request to / should return a valid response"""
        resp_json = self._request_valid("", check_links=False)
        self.assertEqual(resp_json["id"], "eodag-stac-api")
        self.assertEqual(resp_json["title"], "eodag-stac-api")
        self.assertEqual(resp_json["description"], "STAC API provided by EODAG")

        # customize root info
        try:
            Settings.from_environment.cache_clear()
            with temporary_environment(
                EODAG_STAC_API_LANDING_ID="foo-id",
                EODAG_STAC_API_TITLE="foo title",
                EODAG_STAC_API_DESCRIPTION="foo description",
            ):
                resp_json = self._request_valid("", check_links=False)
                self.assertEqual(resp_json["id"], "foo-id")
                self.assertEqual(resp_json["title"], "foo title")
                self.assertEqual(resp_json["description"], "foo description")
        finally:
            Settings.from_environment.cache_clear()

    def test_conformance(self):
        """Request to /conformance should return a valid response"""
        self._request_valid("conformance", check_links=False)

    def test_service_desc(self):
        """Request to service_desc should return a valid response"""
        service_desc = self._request_valid("api", check_links=False)
        self.assertIn("openapi", service_desc.keys())
        self.assertIn("eodag", service_desc["info"]["title"].lower())
        self.assertGreater(len(service_desc["paths"].keys()), 0)
        # test a 2nd call (ending slash must be ignored)
        self._request_valid("api/", check_links=False)

    def test_service_doc(self):
        """Request to service_doc should return a valid response"""
        response = self.app.get("api.html", follow_redirects=True)
        self.assertEqual(200, response.status_code)

    def test_stac_extension_oseo(self):
        """Request to oseo extension should return a valid response"""
        response = self._request_valid(
            "/extensions/oseo/json-schema/schema.json", check_links=False
        )
        self.assertEqual(response["title"], "OpenSearch for Earth Observation")
        self.assertEqual(response["allOf"][0]["$ref"], "#/definitions/oseo")

    def test_queryables(self):
        """Request to /queryables without parameter should return a valid response."""
        stac_common_queryables = list(StacQueryables.default_properties.keys())

        # neither provider nor product type are specified
        res_no_product_type_no_provider = self._request_valid(
            "queryables", check_links=False
        )

        # the response is in StacQueryables class format
        self.assertListEqual(
            list(res_no_product_type_no_provider.keys()),
            [
                "$schema",
                "$id",
                "type",
                "title",
                "description",
                "properties",
                "additionalProperties",
            ],
        )
        self.assertTrue(res_no_product_type_no_provider["additionalProperties"])

        # properties from stac common queryables are added and are the only ones of the response
        self.assertListEqual(
            list(res_no_product_type_no_provider["properties"].keys()),
            stac_common_queryables,
        )

    @mock.patch("eodag.plugins.search.qssearch.requests.Session.get", autospec=True)
    def test_queryables_with_provider(self, mock_requests_get: Mock):
        """Request to /queryables with a valid provider as parameter should return a valid response."""
        queryables_path = os.path.join(
            TEST_RESOURCES_PATH, "stac/provider_queryables.json"
        )
        with open(queryables_path) as f:
            provider_queryables = json.load(f)
        mock_requests_get.return_value = MockResponse(
            provider_queryables, status_code=200
        )

        stac_common_queryables = list(StacQueryables.default_properties.keys())
        provider_stac_queryables_from_queryables_file = [
            "id",
            "gsd",
            "title",
            "s3:gsd",
            "datetime",
            "geometry",
            "platform",
            "processing:level",
            "s1:processing_level",
            "landsat:processing_level",
        ]

        # provider is specified without product type
        res_no_product_type_with_provider = self._request_valid(
            "queryables?provider=planetary_computer", check_links=False
        )

        mock_requests_get.assert_called_once_with(
            mock.ANY,
            url="https://planetarycomputer.microsoft.com/api/stac/v1/search/../queryables",
            timeout=DEFAULT_SEARCH_TIMEOUT,
            headers=USER_AGENT,
            verify=True,
        )

        # the response is in StacQueryables class format
        self.assertListEqual(
            list(res_no_product_type_with_provider.keys()),
            [
                "$schema",
                "$id",
                "type",
                "title",
                "description",
                "properties",
                "additionalProperties",
            ],
        )
        self.assertTrue(res_no_product_type_with_provider["additionalProperties"])

        # properties from stac common queryables are added
        for p in stac_common_queryables:
            self.assertIn(
                p, list(res_no_product_type_with_provider["properties"].keys())
            )

        # properties from provider queryables are added (here the ones of planetary_computer)
        for provider_stac_queryable in provider_stac_queryables_from_queryables_file:
            self.assertIn(
                provider_stac_queryable, res_no_product_type_with_provider["properties"]
            )

    def test_queryables_with_provider_error(self):
        """Request to /queryables with a wrong provider as parameter should return a UnsupportedProvider error."""
        response = self.app.get(
            "queryables?provider=not_supported_provider", follow_redirects=True
        )
        response_content = json.loads(response.content.decode("utf-8"))

        self.assertIn("description", response_content)
        self.assertIn("UnsupportedProvider", response_content["description"])

        self.assertEqual(404, response.status_code)

    @mock.patch("eodag.plugins.manager.PluginManager.get_auth_plugin", autospec=True)
    def test_product_type_queryables(self, mock_requests_session_post):
        """Request to /collections/{collection_id}/queryables should return a valid response."""

        @responses.activate(registry=responses.registries.OrderedRegistry)
        def run():
            queryables_path = os.path.join(
                TEST_RESOURCES_PATH, "stac/product_type_queryables.json"
            )
            with open(queryables_path) as f:
                provider_queryables = json.load(f)
            constraints_path = os.path.join(TEST_RESOURCES_PATH, "constraints.json")
            with open(constraints_path) as f:
                constraints = json.load(f)
            wekeo_main_constraints = {"constraints": constraints}

            planetary_computer_queryables_url = (
                "https://planetarycomputer.microsoft.com/api/stac/v1/collections/"
                "sentinel-1-grd/queryables"
            )
            dedl_queryables_url = (
                "https://hda.data.destination-earth.eu/stac/collections/"
                "EO.ESA.DAT.SENTINEL-1.L1_GRD/queryables"
            )
            wekeo_main_constraints_url = (
                "https://gateway.prod.wekeo2.eu/hda-broker/api/v1/dataaccess/queryable/"
                "EO:ESA:DAT:SENTINEL-1"
            )

            responses.add(
                responses.GET,
                planetary_computer_queryables_url,
                status=200,
                json=provider_queryables,
            )
            responses.add(
                responses.GET,
                wekeo_main_constraints_url,
                status=200,
                json=wekeo_main_constraints,
            )
            responses.add(
                responses.GET,
                dedl_queryables_url,
                status=200,
                json=provider_queryables,
            )

            # no provider is specified with the product type (3 providers get a queryables or constraints file
            # among available providers for S1_SAR_GRD for the moment): queryables intersection returned
            res_product_type_no_provider = self._request_valid(
                "collections/S1_SAR_GRD/queryables",
                check_links=False,
            )
            self.assertEqual(len(responses.calls), 3)

            # check the mock call on planetary_computer
            self.assertEqual(
                planetary_computer_queryables_url, responses.calls[0].request.url
            )
            self.assertIn(
                ("timeout", DEFAULT_SEARCH_TIMEOUT),
                responses.calls[0].request.req_kwargs.items(),
            )
            self.assertIn(
                list(USER_AGENT.items())[0], responses.calls[0].request.headers.items()
            )
            self.assertIn(
                ("verify", True), responses.calls[0].request.req_kwargs.items()
            )
            # check the mock call on wekeo_main
            self.assertEqual(wekeo_main_constraints_url, responses.calls[1].request.url)
            self.assertIn(
                ("timeout", 60), responses.calls[1].request.req_kwargs.items()
            )
            self.assertIn(
                list(USER_AGENT.items())[0], responses.calls[1].request.headers.items()
            )
            self.assertIn(
                ("verify", True), responses.calls[1].request.req_kwargs.items()
            )

            # the response is in StacQueryables class format
            self.assertListEqual(
                list(res_product_type_no_provider.keys()),
                [
                    "$schema",
                    "$id",
                    "type",
                    "title",
                    "description",
                    "properties",
                    "additionalProperties",
                ],
            )
            self.assertTrue(res_product_type_no_provider["additionalProperties"])

            res_list = list(res_product_type_no_provider["properties"].keys())
            expected = [
                "platform",
                "datetime",
                "start_datetime",
                "end_datetime",
                "geometry",
                "bbox",
                "constellation",
                "instruments",
                "gsd",
                "eo:cloud_cover",
                "eo:snow_cover",
                "processing:level",
                "sat:orbit_state",
                "sat:absolute_orbit",
                "sar:instrument_mode",
            ]
            for value in expected:
                self.assertIn(value, res_list)

            # stac format processing:level is in result
            pl_s1_sar_grd_planetary_computer_queryable = "s1:processing_level"
            pl_s1_sar_grd_wekeo_main_queryable = "processingLevel"
            stac_pl_property = "processing:level"
            self.assertIn(
                pl_s1_sar_grd_planetary_computer_queryable,
                provider_queryables["properties"],
            )
            for constraint in wekeo_main_constraints["constraints"]:
                self.assertNotIn(pl_s1_sar_grd_wekeo_main_queryable, constraint)
            self.assertIn(stac_pl_property, res_product_type_no_provider["properties"])

        run()

    def test_product_type_queryables_error(self):
        """Request to /collections/{collection_id}/queryables with a wrong collection_id
        should return a UnsupportedProductType error."""
        response = self.app.get(
            "collections/not_supported_product_type/queryables", follow_redirects=True
        )
        response_content = json.loads(response.content.decode("utf-8"))

        self.assertIn("description", response_content)
        self.assertIn("UnsupportedProductType", response_content["description"])

        self.assertEqual(404, response.status_code)

    @mock.patch("eodag.plugins.search.qssearch.requests.Session.get", autospec=True)
    @mock.patch(
        "eodag.plugins.search.build_search_result.ECMWFSearch.discover_queryables",
        autospec=True,
    )
    def test_product_type_queryables_with_provider(
        self, mock_discover_queryables_ecmwf, mock_requests_get
    ):
        """Request a collection-specific list of queryables for a given provider
        using a queryables file should return a valid response."""
        queryables_path = os.path.join(
            TEST_RESOURCES_PATH, "stac/product_type_queryables.json"
        )
        with open(queryables_path) as f:
            provider_queryables = json.load(f)
        mock_requests_get.return_value = MockResponse(
            provider_queryables, status_code=200
        )

        planetary_computer_queryables_url = (
            "https://planetarycomputer.microsoft.com/api/stac/v1/search/../collections/"
            "sentinel-1-grd/queryables"
        )

        stac_common_queryables = list(StacQueryables.default_properties.keys())
        # when product type is given, "collection" item is not used
        stac_common_queryables.remove("collection")
        provider_stac_queryables_from_queryables_file = [
            "id",
            "datetime",
            "geometry",
            "platform",
            "processing:level",
        ]

        # provider and product type are specified
        res_product_type_with_provider = self._request_valid(
            "collections/S1_SAR_GRD/queryables?provider=planetary_computer",
            check_links=False,
        )

        mock_requests_get.assert_called_once_with(
            mock.ANY,
            url=planetary_computer_queryables_url,
            timeout=DEFAULT_SEARCH_TIMEOUT,
            headers=USER_AGENT,
            verify=True,
        )

        # the response is in StacQueryables class format
        self.assertListEqual(
            list(res_product_type_with_provider.keys()),
            [
                "$schema",
                "$id",
                "type",
                "title",
                "description",
                "properties",
                "additionalProperties",
            ],
        )
        self.assertTrue(res_product_type_with_provider["additionalProperties"])

        # properties from stac common queryables are added
        for p in stac_common_queryables:
            self.assertIn(p, list(res_product_type_with_provider["properties"].keys()))

        # properties from provider product type queryables are added
        # (here the ones of S1_SAR_GRD for planetary_computer)
        for provider_stac_queryable in provider_stac_queryables_from_queryables_file:
            self.assertIn(
                provider_stac_queryable, res_product_type_with_provider["properties"]
            )

        # properties may be updated with info from provider queryables if
        # info exist (here an example with platformSerialIdentifier)
        stac_psi_property = "platform"
        self.assertEqual(
            "string", provider_queryables["properties"][stac_psi_property]["type"]
        )
        self.assertIn(
            "string",
            res_product_type_with_provider["properties"][stac_psi_property]["type"],
        )

        # provider + product type with constraints -> additionalProperties = False
        mock_discover_queryables_ecmwf.return_value = {}
        res_product_type_with_provider = self._request_valid(
            "collections/ERA5_SL/queryables?provider=cop_cds",
            check_links=False,
        )
        self.assertFalse(res_product_type_with_provider["additionalProperties"])

    def test_stac_queryables_type(self):
        res = self._request_valid(
            "collections/S2_MSI_L2A/queryables?provider=creodias",
            check_links=False,
        )
        self.assertIn("eo:cloud_cover", res["properties"])
        cloud_cover = res["properties"]["eo:cloud_cover"]
        self.assertIn("type", cloud_cover)
        self.assertEqual("integer", cloud_cover["type"])
        self.assertIn("exclusiveMinimum", cloud_cover)
        self.assertEqual(0, cloud_cover["exclusiveMinimum"])
        self.assertIn("exclusiveMaximum", cloud_cover)
        self.assertEqual(100, cloud_cover["exclusiveMaximum"])
        self.assertIn("processing:level", res["properties"])
        processing_level = res["properties"]["processing:level"]
        self.assertEqual("string", processing_level["type"])
        self.assertNotIn(
            "min", processing_level
        )  # none values are left out in serialization

    @mock.patch("eodag.utils.requests.requests.sessions.Session.get", autospec=True)
    def test_product_type_queryables_from_constraints(self, mock_requests_get: Mock):
        """Request a collection-specific list of queryables for a given provider
        using a constraints file should return a valid response."""
        constraints_path = os.path.join(TEST_RESOURCES_PATH, "constraints.json")
        with open(constraints_path) as f:
            constraints = json.load(f)
        for const in constraints:
            const["variable"].append("10m_u_component_of_wind")

        form_path = os.path.join(TEST_RESOURCES_PATH, "form.json")
        with open(form_path) as f:
            form = json.load(f)
        mock_requests_get.return_value.json.side_effect = [constraints, form]

        provider_queryables_from_constraints_file = [
            "year",
            "month",
            "day",
            "time",
            "variable",
            "leadtime_hour",
            "type",
            "product_type",
        ]
        # queryables properties not shared by all constraints must be removed
        not_shared_properties = ["leadtime_hour", "type"]
        provider_queryables_from_constraints_file = [
            f"ecmwf:{properties}"
            for properties in provider_queryables_from_constraints_file
            if properties not in not_shared_properties
        ]
        default_provider_stac_properties = ["ecmwf:product_type", "ecmwf:data_format"]

        res = self._request_valid(
            "collections/ERA5_SL/queryables?provider=cop_cds",
            check_links=False,
        )

        mock_requests_get.assert_has_calls(
            [
                call(
                    mock.ANY,
                    "https://cds.climate.copernicus.eu/api/catalogue/v1/collections/"
                    "reanalysis-era5-single-levels/constraints.json",
                    headers=USER_AGENT,
                    auth=mock.ANY,
                    timeout=30,
                ),
                call().raise_for_status(),
                call().json(),
                call(
                    mock.ANY,
                    "https://cds.climate.copernicus.eu/api/catalogue/v1/collections/"
                    "reanalysis-era5-single-levels/form.json",
                    headers=USER_AGENT,
                    auth=mock.ANY,
                    timeout=30,
                ),
                call().raise_for_status(),
                call().json(),
            ]
        )

        # the response is in StacQueryables class format
        self.assertListEqual(
            list(res.keys()),
            [
                "$schema",
                "$id",
                "type",
                "title",
                "description",
                "properties",
                "required",
                "additionalProperties",
            ],
        )
        self.assertFalse(res["additionalProperties"])

        # properties from provider product type queryables and default properties are added
        # (here the ones of ERA5_SL for cop_cds)
        for provider_stac_queryable in list(
            set(
                provider_queryables_from_constraints_file
                + default_provider_stac_properties
            )
        ):
            self.assertIn(provider_stac_queryable, res["properties"])

    def test_cql_post_search(self):
        self._request_valid(
            "search",
            method="POST",
            post_data={
                "filter": {
                    "op": "and",
                    "args": [
                        {
                            "op": "in",
                            "args": [{"property": "id"}, ["foo", "bar"]],
                        },
                        {
                            "op": "=",
                            "args": [
                                {"property": "collection"},
                                self.tested_product_type,
                            ],
                        },
                    ],
                }
            },
            search_call_count=2,
            expected_search_kwargs=[
                {
                    "provider": None,
                    "id": "foo",
                    "productType": self.tested_product_type,
                },
                {
                    "provider": None,
                    "id": "bar",
                    "productType": self.tested_product_type,
                },
            ],
        )

        self._request_valid(
            "search",
            method="POST",
            post_data={
                "filter-lang": "cql2-json",
                "filter": {
                    "op": "and",
                    "args": [
                        {
                            "op": "=",
                            "args": [
                                {"property": "collection"},
                                self.tested_product_type,
                            ],
                        },
                        {"op": "=", "args": [{"property": "eo:cloud_cover"}, 10]},
                        {
                            "op": "t_intersects",
                            "args": [
                                {"property": "datetime"},
                                {
                                    "interval": [
                                        "2018-01-20T00:00:00Z",
                                        "2018-01-25T00:00:00Z",
                                    ]
                                },
                            ],
                        },
                        {
                            "op": "s_intersects",
                            "args": [
                                {"property": "geometry"},
                                {
                                    "type": "Polygon",
                                    "coordinates": [
                                        [[0, 43], [0, 44], [1, 44], [1, 43], [0, 43]]
                                    ],
                                },
                            ],
                        },
                    ],
                },
            },
            expected_search_kwargs={
                "productType": "S2_MSI_L1C",
                "geom": {
                    "type": "Polygon",
                    "coordinates": [[[0, 43], [0, 44], [1, 44], [1, 43], [0, 43]]],
                },
                "start": "2018-01-20T00:00:00Z",
                "end": "2018-01-25T00:00:00Z",
                "cloudCover": 10,
                "page": 1,
                "items_per_page": 20,
                "raise_errors": False,
                "count": True,
            },
        )

        self._request_not_valid(
            "search",
            method="POST",
            post_data={
                "filter": {
                    "op": "and",
                    "args": [
                        {
                            "op": "in",
                            "args": [{"property": "id"}, "foo", "bar"],
                        },
                        {
                            "op": "=",
                            "args": [
                                {"property": "collections"},
                                self.tested_product_type,
                            ],
                        },
                    ],
                }
            },
        )

    @mock.patch("eodag.rest.core.eodag_api.list_product_types", autospec=True)
    @mock.patch("eodag.rest.core.eodag_api.guess_product_type", autospec=True)
    def test_collection_free_text_search(self, guess_pt: Mock, list_pt: Mock):
        """Test STAC Collection free-text search"""

        url = "/collections?q=TERM1,TERM2"
        r = self.app.get(url)
        list_pt.assert_called_once_with(provider=None, fetch_providers=False)
        guess_pt.assert_called_once_with(
            free_text="TERM1,TERM2",
            platformSerialIdentifier=None,
            instrument=None,
            platform=None,
            missionStartDate=None,
            missionEndDate=None,
            productType=None,
        )
        self.assertEqual(200, r.status_code)
